'''
High level ipset support.

When :doc:`ipset` is providing a direct netlink socket with low level
functions, a :class:`WiSet` object is built to map ipset objects from kernel.
It helps to add/remove entries, list content, etc.

For example, adding an entry with :class:`pyroute2.ipset.IPSet` object
implies to set a various number of parameters:

.. doctest::
    :skipif: True

    >>> ipset = IPSet()
    >>> ipset.add("foo", "1.2.3.4/24", etype="net")
    >>> ipset.close()

When they are discovered by a :class:`WiSet`:

.. doctest::
    :skipif: True

    >>> wiset = load_ipset("foo")
    >>> wiset.add("1.2.3.4/24")

Listing entries is also easier using :class:`WiSet`, since it parses for you
netlink messages:

.. doctest::
    :skipif: True

    >>> wiset.content
    {'1.2.3.0/24': IPStats(packets=None, bytes=None, comment=None,
                           timeout=None, skbmark=None, physdev=False)}
'''

import errno
import uuid
from collections import namedtuple
from inspect import getcallargs
from socket import AF_INET

from pyroute2.common import basestring
from pyroute2.ipset import IPSet
from pyroute2.netlink.exceptions import IPSetError
from pyroute2.netlink.nfnetlink.ipset import (
    IPSET_FLAG_IFACE_WILDCARD,
    IPSET_FLAG_PHYSDEV,
    IPSET_FLAG_WITH_COMMENT,
    IPSET_FLAG_WITH_COUNTERS,
    IPSET_FLAG_WITH_SKBINFO,
)
from pyroute2.netlink.nfnetlink.nfctsocket import IP_PROTOCOLS

# Debug variable to detect netlink socket leaks
COUNT = {"count": 0}


def need_ipset_socket(fun):
    """Decorator to create netlink socket if needed.

    In many of our helpers, we need to open a netlink socket. This can
    be expensive for someone using many times the functions: instead to have
    only one socket and use several requests, we will open it again and again.

    This helper allow our functions to be flexible: the caller can pass an
    optional socket, or do nothing. In this last case, this decorator
    will open a socket for the caller (and close it after call)

    It also help to mix helpers. One helper can call another one: the socket
    will be opened only once. We just have to pass the ipset variable.

    Note that all functions using this helper *must* use ipset as variable
    name for the socket.
    """

    def wrap(*args, **kwargs):
        callargs = getcallargs(fun, *args, **kwargs)
        if callargs["sock"] is None:
            # This variable is used only to debug leak in tests
            COUNT['count'] += 1
            with IPSet() as sock:
                callargs["sock"] = sock
                # We must pop kwargs here, else the function will receive
                # a dict of dict
                if "kwargs" in callargs:
                    callargs.update(callargs.pop("kwargs"))
                return fun(**callargs)  # pylint:disable=star-args

        return fun(*args, **kwargs)

    return wrap


class IPStats(
    namedtuple(
        "IPStats",
        [
            "packets",
            "bytes",
            "comment",
            "timeout",
            "skbmark",
            "physdev",
            "wildcard",
        ],
    )
):
    __slots__ = ()

    def __new__(
        cls,
        packets,
        bytes,
        comment,
        timeout,
        skbmark,
        physdev=False,
        wildcard=False,
    ):
        return super(IPStats, cls).__new__(
            cls,
            packets,
            bytes,
            comment,
            timeout,
            skbmark,
            physdev=physdev,
            wildcard=wildcard,
        )


# pylint: disable=too-many-instance-attributes
class WiSet(object):
    """Main high level ipset manipulation class.

    Every high level ipset operation should be possible with this class,
    you probably don't need other helpers of this module, except tools
    to load data from kernel (:func:`load_all_ipsets` and :func:`load_ipset`)

    For example, you can create and an entry in a ipset just with:

    .. doctest::
        :skipif: True

        >>> with WiSet(name="mysuperipset") as myset:
        >>>    myset.create()             # add the ipset in the kernel
        >>>    myset.add("198.51.100.1")  # add one IP to the set

    Netlink sockets are opened by __enter__ and __exit__ function, so you don't
    have to manage it manually if you use the "with" keyword.

    If you want to manage it manually (for example for long operation in
    a daemon), you can do the following:

    .. doctest::
        :skipif: True

        >>> myset = WiSet(name="mysuperipset")
        >>> myset.open_netlink()
        >>> # do stuff
        >>> myset.close_netlink()

    You can also don't initiate at all any netlink socket, this code will work:

    .. doctest::
        :skipif: True

        >>> myset = WiSet(name="mysuperipset")
        >>> myset.create()
        >>> myset.destroy()

    But do it very carefully. In that case, a netlink socket will be opened
    in background for any operation. No socket will be leaked, but that
    can consume resources.

    You can also instantiate WiSet objects with :func:`load_all_ipsets` and
    :func:`load_ipset`:

    .. doctest::
        :skipif: True

        >>> all_sets_dict = load_all_ipsets()
        >>> one_set = load_ipset(name="myset")

    Have a look on content variable if you need list of entries in the Set.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name=None,
        attr_type='hash:ip',
        family=AF_INET,
        sock=None,
        timeout=None,
        counters=False,
        comment=False,
        hashsize=None,
        revision=None,
        skbinfo=False,
    ):
        self.name = name
        self.hashsize = hashsize
        self._attr_type = None
        self.entry_type = None
        self.attr_type = attr_type
        self.family = family
        self._content = None
        self.sock = sock
        self.timeout = timeout
        self.counters = counters
        self.comment = comment
        self.revision = revision
        self.index = None
        self.skbinfo = skbinfo

    def open_netlink(self):
        """
        Open manually a netlink socket.

        You can use "with WiSet()" statement instead.
        """
        if self.sock is None:
            self.sock = IPSet()

    def close_netlink(self):
        """Clone any opened netlink socket"""
        if self.sock is not None:
            self.sock.close()
            self.sock = None

    @property
    def attr_type(self):
        return self._attr_type

    @attr_type.setter
    def attr_type(self, value):
        self._attr_type = value
        self.entry_type = value.split(":", 1)[1]

    def __enter__(self):
        self.open_netlink()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_netlink()

    @classmethod
    def from_netlink(cls, ndmsg, content=False):
        """Create a ipset objects based on a parsed netlink message

        :param ndmsg: the netlink message to parse
        :param content: should we fill (and parse) entries info (can be slow
                        on very large set)
        :type content: bool
        """
        self = cls()
        self.attr_type = ndmsg.get_attr("IPSET_ATTR_TYPENAME")
        self.name = ndmsg.get_attr("IPSET_ATTR_SETNAME")
        self.hashsize = ndmsg.get_attr("IPSET_ATTR_HASHSIZE")
        self.family = ndmsg.get_attr("IPSET_ATTR_FAMILY")
        self.revision = ndmsg.get_attr("IPSET_ATTR_REVISION")
        self.index = ndmsg.get_attr("IPSET_ATTR_INDEX")
        data = ndmsg.get_attr("IPSET_ATTR_DATA")
        self.timeout = data.get_attr("IPSET_ATTR_TIMEOUT")
        flags = data.get_attr("IPSET_ATTR_CADT_FLAGS")
        if flags is not None:
            self.counters = bool(flags & IPSET_FLAG_WITH_COUNTERS)
            self.comment = bool(flags & IPSET_FLAG_WITH_COMMENT)
            self.skbinfo = bool(flags & IPSET_FLAG_WITH_SKBINFO)

        if content:
            self.update_dict_content(ndmsg)

        return self

    def update_dict_content(self, ndmsg):
        """Update a dictionary statistics with values sent in netlink message

        :param ndmsg: the netlink message
        :type ndmsg: netlink message

        """
        family = "IPSET_ATTR_IPADDR_IPV4"
        ip_attr = "IPSET_ATTR_IP_FROM"
        if self._content is None:
            self._content = {}

        timeout = None
        entries = ndmsg.get_attr("IPSET_ATTR_ADT").get_attrs("IPSET_ATTR_DATA")
        for entry in entries:
            key = ""
            for parse_type in self.entry_type.split(","):
                if parse_type == "ip":
                    ip = entry.get_attr(ip_attr).get_attr(family)
                    key += ip
                elif parse_type == "net":
                    ip = entry.get_attr(ip_attr).get_attr(family)
                    key += ip
                    cidr = entry.get_attr("IPSET_ATTR_CIDR")
                    if cidr is not None:
                        key += "/{0}".format(cidr)
                elif parse_type == "iface":
                    key += entry.get_attr("IPSET_ATTR_IFACE")
                elif parse_type == "set":
                    key += entry.get_attr("IPSET_ATTR_NAME")
                elif parse_type == "mark":
                    key += str(hex(entry.get_attr("IPSET_ATTR_MARK")))
                elif parse_type == "port":
                    proto = entry.get_attr('IPSET_ATTR_PROTO')
                    if proto is not None:
                        proto = IP_PROTOCOLS.get(proto, str(proto)).lower()
                        key += '{proto}:'.format(proto=proto)
                    key += str(entry.get_attr("IPSET_ATTR_PORT_FROM"))
                elif parse_type == "mac":
                    key += entry.get_attr("IPSET_ATTR_ETHER")
                key += ","

            key = key.strip(",")

            if self.timeout is not None:
                timeout = entry.get_attr("IPSET_ATTR_TIMEOUT")
            skbmark = entry.get_attr("IPSET_ATTR_SKBMARK")
            if skbmark is not None:
                # Convert integer to hex for mark/mask
                # Only display mask if != 0xffffffff
                if skbmark[1] != (2**32 - 1):
                    skbmark = "/".join([str(hex(mark)) for mark in skbmark])
                else:
                    skbmark = str(hex(skbmark[0]))
            entry_flag_parsed = {"physdev": False}
            flags = entry.get_attr("IPSET_ATTR_CADT_FLAGS")
            if flags is not None:
                entry_flag_parsed["physdev"] = bool(flags & IPSET_FLAG_PHYSDEV)
                entry_flag_parsed["wildcard"] = bool(
                    flags & IPSET_FLAG_IFACE_WILDCARD
                )

            value = IPStats(
                packets=entry.get_attr("IPSET_ATTR_PACKETS"),
                bytes=entry.get_attr("IPSET_ATTR_BYTES"),
                comment=entry.get_attr("IPSET_ATTR_COMMENT"),
                skbmark=skbmark,
                timeout=timeout,
                **entry_flag_parsed
            )
            self._content[key] = value

    def create(self, **kwargs):
        """Insert this Set in the kernel

        Many options are set with python object attributes (like comments,
        counters, etc). For non-supported type, kwargs are provided. See
        :doc:`ipset` documentation for more information.
        """
        create_ipset(
            self.name,
            stype=self.attr_type,
            family=self.family,
            sock=self.sock,
            timeout=self.timeout,
            comment=self.comment,
            counters=self.counters,
            hashsize=self.hashsize,
            skbinfo=self.skbinfo,
            **kwargs
        )

    def destroy(self):
        """Destroy this ipset in the kernel list.

        It does not delete this python object (any content or other stored
        values are keep in memory). This function will fail if the ipset is
        still referenced (by example in iptables rules), you have been warned.
        """
        destroy_ipset(self.name, sock=self.sock)

    def add(self, entry, **kwargs):
        """Add an entry in this ipset.

        If counters are enabled on the set, reset by default the value when
        we add the element. Without this reset, kernel sometimes store old
        values and can add very strange behavior on counters.
        """
        if isinstance(entry, dict):
            kwargs.update(entry)
            entry = kwargs.pop("entry")
        if self.counters:
            kwargs["packets"] = kwargs.pop("packets", 0)
            kwargs["bytes"] = kwargs.pop("bytes", 0)
        skbmark = kwargs.get("skbmark")
        if isinstance(skbmark, basestring):
            skbmark = skbmark.split('/')
            mark = int(skbmark[0], 16)
            try:
                mask = int(skbmark[1], 16)
            except IndexError:
                mask = int("0xffffffff", 16)
            kwargs["skbmark"] = (mark, mask)
        add_ipset_entry(
            self.name, entry, etype=self.entry_type, sock=self.sock, **kwargs
        )

    def delete(self, entry, **kwargs):
        """Delete/remove an entry in this ipset"""
        delete_ipset_entry(
            self.name, entry, etype=self.entry_type, sock=self.sock, **kwargs
        )

    def test(self, entry, **kwargs):
        """Test if an entry is in this ipset"""
        return test_ipset_entry(
            self.name, entry, etype=self.entry_type, sock=self.sock, **kwargs
        )

    def test_list(self, entries, **kwargs):
        """Test if a list of a set of entries is in this ipset

        Return a set of entries found in the IPSet
        """
        return test_ipset_entries(
            self.name, entries, etype=self.entry_type, sock=self.sock, **kwargs
        )

    def update_content(self):
        """Update the content dictionary with values from kernel"""
        self._content = {}
        update_wiset_content(self, sock=self.sock)

    def flush(self):
        """Flush entries of the ipset"""
        flush_ipset(self.name, sock=self.sock)

    @property
    def content(self):
        """Dictionary of entries in the set.

        Keys are IP addresses (as string), values are IPStats tuples.
        """
        if self._content is None:
            self.update_content()

        return self._content

    def insert_list(self, entries):
        """Just a small helper to reduce the number of loops in main code."""
        for entry in entries:
            self.add(entry)

    def replace_entries(self, new_list):
        """Replace the content of an ipset with a new list of entries.

        This operation is like a flush() and adding all entries one by one. But
        this call is atomic: it creates a temporary ipset and swap the content.

        :param new_list: list of entries to add
        :type new_list: list or :py:class:`set` of basestring or of
            keyword arguments dict
        """
        temp_name = str(uuid.uuid4())[0:8]
        # Get a copy of ourself
        temp = load_ipset(self.name, sock=self.sock)
        temp.name = temp_name
        temp.sock = self.sock
        temp.create()
        temp.insert_list(new_list)
        swap_ipsets(self.name, temp_name, sock=self.sock)
        temp.destroy()


@need_ipset_socket
def create_ipset(
    name, stype=None, family=AF_INET, exclusive=False, sock=None, **kwargs
):
    """Create an ipset."""
    sock.create(
        name, stype=stype, family=family, exclusive=exclusive, **kwargs
    )


@need_ipset_socket
def load_all_ipsets(content=False, sock=None, inherit_sock=False, prefix=None):
    """List all ipset as WiSet objects.

    Get full ipset data from kernel and parse it in WiSet objects. Result is
    a dictionary with ipset names as keys, and WiSet objects as values.

    :param content: parse the list of entries and fill it in WiSet content
                    dictionary
    :type content: bool
    :param inherit_sock: use the netlink sock passed in ipset arg to
                         fill WiSets sock
    :type inherit_sock: bool
    :param prefix: filter out all ipset with a name not beginning by this
                   prefix
    :type prefix: str or None
    """
    res = {}
    for myset in sock.list():
        # on large sets, we can receive data in several messages
        name = myset.get_attr("IPSET_ATTR_SETNAME")
        if prefix is not None and not name.startswith(prefix):
            continue
        if name not in res:
            wiset = WiSet.from_netlink(myset, content=content)
            if inherit_sock:
                wiset.sock = sock
            res[wiset.name] = wiset
        elif content:
            res[wiset.name].update_dict_content(myset)
    return res


@need_ipset_socket
def load_ipset(name, content=False, sock=None, inherit_sock=False):
    """Get one ipset as WiSet object

    Helper to get current WiSet object. More efficient that
    :func:`load_all_ipsets` since the kernel does the filtering itself.

    Return None if the ipset does not exist

    :param name: name of the ipset
    :type name: str
    :param content: parse or not content and statistics on entries
    :type content: bool
    :param inherit_sock: use the netlink sock passed in ipset arg to
                         fill WiSet sock
    :type inherit_sock: bool
    """
    res = None
    try:
        messages = sock.list(name=name)
    except IPSetError as e:
        if e.code == errno.ENOENT:
            return res
        raise
    for msg in messages:
        if res is None:
            res = WiSet.from_netlink(msg, content=content)
            if inherit_sock:
                res.sock = sock
        elif content:
            res.update_dict_content(msg)
    return res


@need_ipset_socket
def update_wiset_content(wiset, sock=None):
    """Update content/statistics of a wiset.

    You should never call yourself this function. It is only a helper to use
    the :func:`need_ipset_socket` decorator out of WiSet object.
    """
    for msg in sock.list(name=wiset.name):
        wiset.update_dict_content(msg)


@need_ipset_socket
def destroy_ipset(name, sock=None):
    """Remove an ipset in the kernel."""
    sock.destroy(name)


@need_ipset_socket
def add_ipset_entry(name, entry, sock=None, **kwargs):
    """Add an entry"""
    sock.add(name, entry, **kwargs)


@need_ipset_socket
def delete_ipset_entry(name, entry, sock=None, **kwargs):
    """Remove one entry"""
    sock.delete(name, entry, **kwargs)


@need_ipset_socket
def test_ipset_exist(name, sock=None):
    """Test if the given ipset exist"""
    try:
        sock.headers(name)
        return True
    except IPSetError as e:
        if e.code == errno.ENOENT:
            return False
        raise


@need_ipset_socket
def test_ipset_entry(name, entry, sock=None, **kwargs):
    """Test if an entry is in one ipset"""
    return sock.test(name, entry, **kwargs)


@need_ipset_socket
def test_ipset_entries(name, entries, sock=None, **kwargs):
    """Test a list (or a set) of entries."""
    res = set()
    for entry in entries:
        if sock.test(name, entry, **kwargs):
            res.add(entry)
    return res


@need_ipset_socket
def flush_ipset(name, sock=None):
    """Flush all ipset content"""
    sock.flush(name)


@need_ipset_socket
def swap_ipsets(name_a, name_b, sock=None):
    """Swap the content of ipset a and b.

    ipsets must have compatible content.
    """
    sock.swap(name_a, name_b)


def get_ipset_socket(**kwargs):
    """Get a socket that one can pass to several WiSet objects"""
    return IPSet(**kwargs)



from pyroute2.netlink import NLA_F_NESTED, NLA_F_NET_BYTEORDER, nla
from pyroute2.netlink.nfnetlink import NFNL_SUBSYS_IPSET, nfgen_msg

IPSET_MAXNAMELEN = 32
IPSET_DEFAULT_MAXELEM = 65536

IPSET_CMD_NONE = 0
IPSET_CMD_PROTOCOL = 1  # Return protocol version
IPSET_CMD_CREATE = 2  # Create a new (empty) set
IPSET_CMD_DESTROY = 3  # Destroy a (empty) set
IPSET_CMD_FLUSH = 4  # Remove all elements from a set
IPSET_CMD_RENAME = 5  # Rename a set
IPSET_CMD_SWAP = 6  # Swap two sets
IPSET_CMD_LIST = 7  # List sets
IPSET_CMD_SAVE = 8  # Save sets
IPSET_CMD_ADD = 9  # Add an element to a set
IPSET_CMD_DEL = 10  # Delete an element from a set
IPSET_CMD_TEST = 11  # Test an element in a set
IPSET_CMD_HEADER = 12  # Get set header data only
IPSET_CMD_TYPE = 13  # 13: Get set type
IPSET_CMD_GET_BYNAME = 14  # 14: Get set index by name
IPSET_CMD_GET_BYINDEX = 15  # 15: Get set index by index

# flags at command level (IPSET_ATTR_FLAGS)
IPSET_FLAG_LIST_SETNAME = 1 << 1
IPSET_FLAG_LIST_HEADER = 1 << 2
IPSET_FLAG_SKIP_COUNTER_UPDATE = 1 << 3
IPSET_FLAG_SKIP_SUBCOUNTER_UPDATE = 1 << 4
IPSET_FLAG_MATCH_COUNTERS = 1 << 5
IPSET_FLAG_RETURN_NOMATCH = 1 << 7

# flags at cadt attribute (IPSET_ATTR_CADT_FLAGS)
IPSET_FLAG_PHYSDEV = 1 << 1
IPSET_FLAG_NOMATCH = 1 << 2
IPSET_FLAG_WITH_COUNTERS = 1 << 3
IPSET_FLAG_WITH_COMMENT = 1 << 4
IPSET_FLAG_WITH_FORCEADD = 1 << 5
IPSET_FLAG_WITH_SKBINFO = 1 << 6
IPSET_FLAG_IFACE_WILDCARD = 1 << 7

IPSET_ERR_PROTOCOL = 4097
IPSET_ERR_FIND_TYPE = 4098
IPSET_ERR_MAX_SETS = 4099
IPSET_ERR_BUSY = 4100
IPSET_ERR_EXIST_SETNAME2 = 4101
IPSET_ERR_TYPE_MISMATCH = 4102
IPSET_ERR_EXIST = 4103
IPSET_ERR_INVALID_CIDR = 4104
IPSET_ERR_INVALID_NETMASK = 4105
IPSET_ERR_INVALID_FAMILY = 4106
IPSET_ERR_TIMEOUT = 4107
IPSET_ERR_REFERENCED = 4108
IPSET_ERR_IPADDR_IPV4 = 4109
IPSET_ERR_IPADDR_IPV6 = 4110
IPSET_ERR_COUNTER = 4111
IPSET_ERR_COMMENT = 4112
IPSET_ERR_INVALID_MARKMASK = 4113
IPSET_ERR_SKBINFO = 4114
IPSET_ERR_TYPE_SPECIFIC = 4352


class ipset_base(nla):
    class ipset_ip(nla):
        nla_flags = NLA_F_NESTED
        nla_map = (
            ('IPSET_ATTR_UNSPEC', 'none'),
            ('IPSET_ATTR_IPADDR_IPV4', 'ip4addr', NLA_F_NET_BYTEORDER),
            ('IPSET_ATTR_IPADDR_IPV6', 'ip6addr', NLA_F_NET_BYTEORDER),
        )


class ipset_msg(nfgen_msg):
    '''
    Since the support just begins to be developed,
    many attrs are still in `hex` format -- just to
    dump the content.
    '''

    nla_map = (
        ('IPSET_ATTR_UNSPEC', 'none'),
        ('IPSET_ATTR_PROTOCOL', 'uint8'),
        ('IPSET_ATTR_SETNAME', 'asciiz'),
        ('IPSET_ATTR_TYPENAME', 'asciiz'),
        ('IPSET_ATTR_REVISION', 'uint8'),
        ('IPSET_ATTR_FAMILY', 'uint8'),
        ('IPSET_ATTR_FLAGS', 'be32'),
        ('IPSET_ATTR_DATA', 'get_data_type'),
        ('IPSET_ATTR_ADT', 'attr_adt'),
        ('IPSET_ATTR_LINENO', 'hex'),
        ('IPSET_ATTR_PROTOCOL_MIN', 'uint8'),
        ('IPSET_ATTR_INDEX', 'be16'),
    )

    @staticmethod
    def get_data_type(self, *args, **kwargs):
        # create and list commands have specific attributes. See linux_ip_set.h
        # for more information (and/or lib/PROTOCOL of ipset sources)
        cmd = self['header']['type'] & ~(NFNL_SUBSYS_IPSET << 8)
        if cmd == IPSET_CMD_CREATE or cmd == IPSET_CMD_LIST:
            return self.cadt_data

        return self.ipset_generic.adt_data

    class ipset_generic(ipset_base):
        class adt_data(ipset_base):
            nla_flags = NLA_F_NESTED
            nla_map = (
                (0, 'IPSET_ATTR_UNSPEC', 'none'),
                (1, 'IPSET_ATTR_IP', 'ipset_ip'),
                (1, 'IPSET_ATTR_IP_FROM', 'ipset_ip'),
                (2, 'IPSET_ATTR_IP_TO', 'ipset_ip'),
                (3, 'IPSET_ATTR_CIDR', 'be8', NLA_F_NET_BYTEORDER),
                (4, 'IPSET_ATTR_PORT', 'be16', NLA_F_NET_BYTEORDER),
                (4, 'IPSET_ATTR_PORT_FROM', 'be16', NLA_F_NET_BYTEORDER),
                (5, 'IPSET_ATTR_PORT_TO', 'be16', NLA_F_NET_BYTEORDER),
                (6, 'IPSET_ATTR_TIMEOUT', 'be32', NLA_F_NET_BYTEORDER),
                (7, 'IPSET_ATTR_PROTO', 'be8', NLA_F_NET_BYTEORDER),
                (8, 'IPSET_ATTR_CADT_FLAGS', 'be32', NLA_F_NET_BYTEORDER),
                (9, 'IPSET_ATTR_CADT_LINENO', 'be32'),
                (10, 'IPSET_ATTR_MARK', 'be32', NLA_F_NET_BYTEORDER),
                (11, 'IPSET_ATTR_MARKMASK', 'be32', NLA_F_NET_BYTEORDER),
                (17, 'IPSET_ATTR_ETHER', 'l2addr'),
                (18, 'IPSET_ATTR_NAME', 'asciiz'),
                (19, 'IPSET_ATTR_NAMEREF', 'be32'),
                (20, 'IPSET_ATTR_IP2', 'ipset_ip'),
                (21, 'IPSET_ATTR_CIDR2', 'be8', NLA_F_NET_BYTEORDER),
                (22, 'IPSET_ATTR_IP2_TO', 'ipset_ip'),
                (23, 'IPSET_ATTR_IFACE', 'asciiz'),
                (24, 'IPSET_ATTR_BYTES', 'be64', NLA_F_NET_BYTEORDER),
                (25, 'IPSET_ATTR_PACKETS', 'be64', NLA_F_NET_BYTEORDER),
                (26, 'IPSET_ATTR_COMMENT', 'asciiz'),
                (27, 'IPSET_ATTR_SKBMARK', 'skbmark'),
                (28, 'IPSET_ATTR_SKBPRIO', 'skbprio'),
                (29, 'IPSET_ATTR_SKBQUEUE', 'be16', NLA_F_NET_BYTEORDER),
            )

            class skbmark(nla):
                nla_flags = NLA_F_NET_BYTEORDER
                fields = [('value', '>II')]

            class skbprio(nla):
                nla_flags = NLA_F_NET_BYTEORDER
                fields = [('value', '>HH')]

    class cadt_data(ipset_base):
        nla_flags = NLA_F_NESTED
        nla_map = (
            (0, 'IPSET_ATTR_UNSPEC', 'none'),
            (1, 'IPSET_ATTR_IP', 'ipset_ip'),
            (1, 'IPSET_ATTR_IP_FROM', 'ipset_ip'),
            (2, 'IPSET_ATTR_IP_TO', 'ipset_ip'),
            (3, 'IPSET_ATTR_CIDR', 'be8', NLA_F_NET_BYTEORDER),
            (4, 'IPSET_ATTR_PORT', 'be16', NLA_F_NET_BYTEORDER),
            (4, 'IPSET_ATTR_PORT_FROM', 'be16', NLA_F_NET_BYTEORDER),
            (5, 'IPSET_ATTR_PORT_TO', 'be16', NLA_F_NET_BYTEORDER),
            (6, 'IPSET_ATTR_TIMEOUT', 'be32', NLA_F_NET_BYTEORDER),
            (7, 'IPSET_ATTR_PROTO', 'be8', NLA_F_NET_BYTEORDER),
            (8, 'IPSET_ATTR_CADT_FLAGS', 'be32', NLA_F_NET_BYTEORDER),
            (9, 'IPSET_ATTR_CADT_LINENO', 'be32'),
            (10, 'IPSET_ATTR_MARK', 'be32', NLA_F_NET_BYTEORDER),
            (11, 'IPSET_ATTR_MARKMASK', 'be32', NLA_F_NET_BYTEORDER),
            (17, 'IPSET_ATTR_INITVAL', 'be32', NLA_F_NET_BYTEORDER),
            (18, 'IPSET_ATTR_HASHSIZE', 'be32', NLA_F_NET_BYTEORDER),
            (19, 'IPSET_ATTR_MAXELEM', 'be32', NLA_F_NET_BYTEORDER),
            (20, 'IPSET_ATTR_NETMASK', 'hex'),
            (21, 'IPSET_ATTR_BUCKETSIZE', 'uint8'),
            (22, 'IPSET_ATTR_RESIZE', 'hex'),
            (23, 'IPSET_ATTR_SIZE', 'be32', NLA_F_NET_BYTEORDER),
            (24, 'IPSET_ATTR_ELEMENTS', 'be32', NLA_F_NET_BYTEORDER),
            (25, 'IPSET_ATTR_REFERENCES', 'be32', NLA_F_NET_BYTEORDER),
            (26, 'IPSET_ATTR_MEMSIZE', 'be32', NLA_F_NET_BYTEORDER),
        )

    class attr_adt(ipset_generic):
        nla_flags = NLA_F_NESTED
        nla_map = ((7, 'IPSET_ATTR_DATA', 'adt_data'),)



from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import TC_H_ROOT
from pyroute2.netlink.rtnl.tcmsg.common import (
    get_size,
    red_eval_ewma,
    red_eval_P,
)

parent = TC_H_ROOT

TC_RED_ECN = 1
TC_RED_HARDDROP = 2
TC_RED_ADAPTATIVE = 4


def get_parameters(kwarg):
    kwarg['quantum'] = get_size(kwarg.get('quantum', 0))
    kwarg['perturb_period'] = kwarg.get('perturb', 0) or kwarg.get(
        'perturb_period', 0
    )
    limit = kwarg['limit'] = kwarg.get('limit', 0) or kwarg.get(
        'redflowlimit', 0
    )
    qth_min = kwarg.get('min', 0)
    qth_max = kwarg.get('max', 0)
    avpkt = kwarg.get('avpkt', 1000)
    probability = kwarg.get('probability', 0.02)
    ecn = kwarg.get('ecn', False)
    harddrop = kwarg.get('harddrop', False)
    kwarg['flags'] = kwarg.get('flags', 0)
    if ecn:
        kwarg['flags'] |= TC_RED_ECN
    if harddrop:
        kwarg['flags'] |= TC_RED_HARDDROP
    if kwarg.get('redflowlimit'):
        qth_max = qth_max or limit / 4
        qth_min = qth_min or qth_max / 3
        kwarg['burst'] = kwarg['burst'] or (2 * qth_min + qth_max) / (
            3 * avpkt
        )
        if limit <= qth_max:
            raise ValueError('limit must be > qth_max')
        if qth_max <= qth_min:
            raise ValueError('qth_max must be > qth_min')
        kwarg['qth_min'] = qth_min
        kwarg['qth_max'] = qth_max
        kwarg['Wlog'] = red_eval_ewma(qth_min, kwarg['burst'], avpkt)
        kwarg['Plog'] = red_eval_P(qth_min, qth_max, probability)
        if kwarg['Wlog'] < 0:
            raise ValueError('Wlog must be > 0')
        if kwarg['Plog'] < 0:
            raise ValueError('Plog must be > 0')
        kwarg['max_P'] = int(probability * pow(2, 23))

    return kwarg


class options_sfq_v0(nla):
    fields = (
        ('quantum', 'I'),
        ('perturb_period', 'i'),
        ('limit', 'I'),
        ('divisor', 'I'),
        ('flows', 'I'),
    )


class options_sfq_v1(nla):
    fields = (
        ('quantum', 'I'),
        ('perturb_period', 'i'),
        ('limit_v0', 'I'),
        ('divisor', 'I'),
        ('flows', 'I'),
        ('depth', 'I'),
        ('headdrop', 'I'),
        ('limit_v1', 'I'),
        ('qth_min', 'I'),
        ('qth_max', 'I'),
        ('Wlog', 'B'),
        ('Plog', 'B'),
        ('Scell_log', 'B'),
        ('flags', 'B'),
        ('max_P', 'I'),
        ('prob_drop', 'I'),
        ('forced_drop', 'I'),
        ('prob_mark', 'I'),
        ('forced_mark', 'I'),
        ('prob_mark_head', 'I'),
        ('forced_mark_head', 'I'),
    )


def options(*argv, **kwarg):
    if kwarg.get('length', 0) >= options_sfq_v1.get_size():
        return options_sfq_v1
    else:
        return options_sfq_v0



import logging
import threading
import traceback
from collections import namedtuple
from socket import AF_INET, AF_INET6

from pyroute2.ipdb.exceptions import CommitException
from pyroute2.ipdb.transactional import Transactional
from pyroute2.netlink import rtnl
from pyroute2.netlink.rtnl.fibmsg import FR_ACT_NAMES, fibmsg

log = logging.getLogger(__name__)
groups = rtnl.RTMGRP_IPV4_RULE | rtnl.RTMGRP_IPV6_RULE


RuleKey = namedtuple(
    'RuleKey',
    (
        'action',
        'table',
        'priority',
        'iifname',
        'oifname',
        'fwmark',
        'fwmask',
        'family',
        'goto',
        'tun_id',
    ),
)


class Rule(Transactional):
    '''
    Persistent transactional rule object
    '''

    _fields = [fibmsg.nla2name(i[1]) for i in fibmsg.nla_map]
    for key, _ in fibmsg.fields:
        _fields.append(key)
    _fields.append('removal')
    _virtual_fields = ['ipdb_scope', 'ipdb_priority']
    _fields.extend(_virtual_fields)
    cleanup = (
        'attrs',
        'header',
        'event',
        'src_len',
        'dst_len',
        'res1',
        'res2',
    )

    @classmethod
    def make_key(cls, msg):
        values = []
        if isinstance(msg, fibmsg):
            for field in RuleKey._fields:
                v = msg.get_attr(msg.name2nla(field))
                if v is None:
                    v = msg.get(field, 0)
                values.append(v)
        elif isinstance(msg, dict):
            for field in RuleKey._fields:
                values.append(msg.get(field, 0))
        else:
            raise TypeError('prime not supported: %s' % type(msg))
        return RuleKey(*values)

    def __init__(self, ipdb, mode=None, parent=None, uid=None):
        Transactional.__init__(self, ipdb, mode, parent, uid)
        with self._direct_state:
            self['ipdb_priority'] = 0

    def load_netlink(self, msg):
        with self._direct_state:
            if self['ipdb_scope'] == 'locked':
                # do not touch locked interfaces
                return

            self['ipdb_scope'] = 'system'
            for key, value in msg.items():
                self[key] = value

            # merge NLA
            for cell in msg['attrs']:
                #
                # Parse on demand
                #
                norm = fibmsg.nla2name(cell[0])
                if norm in self.cleanup:
                    continue
                self[norm] = cell[1]

            if msg.get_attr('FRA_DST'):
                dst = '%s/%s' % (msg.get_attr('FRA_DST'), msg['dst_len'])
                self['dst'] = dst
            if msg.get_attr('FRA_SRC'):
                src = '%s/%s' % (msg.get_attr('FRA_SRC'), msg['src_len'])
                self['src'] = src

            # finally, cleanup all not needed
            for item in self.cleanup:
                if item in self:
                    del self[item]
        return self

    def commit(
        self, tid=None, transaction=None, commit_phase=1, commit_mask=0xFF
    ):
        if not commit_phase & commit_mask:
            return self

        error = None
        drop = self.ipdb.txdrop
        devop = 'set'
        debug = {'traceback': None, 'next_stage': None}
        notx = True

        if tid or transaction:
            notx = False
        if tid:
            transaction = self.global_tx[tid]
        else:
            transaction = transaction or self.current_tx

        # create a new route
        if self['ipdb_scope'] != 'system':
            devop = 'add'

        # work on an existing route
        snapshot = self.pick()
        added, removed = transaction // snapshot
        added.pop('ipdb_scope', None)
        removed.pop('ipdb_scope', None)

        try:
            # rule add/set
            if any(added.values()) or devop == 'add':
                old_key = self.make_key(self)
                new_key = self.make_key(transaction)

                if new_key != old_key:
                    # check for the key conflict
                    if new_key in self.ipdb.rules:
                        raise CommitException('rule priority conflict')
                    else:
                        self.ipdb.rules[new_key] = self
                        self.nl.rule('del', **old_key._asdict())
                        self.nl.rule('add', **transaction)
                else:
                    if devop != 'add':
                        with self._direct_state:
                            self['ipdb_scope'] = 'locked'
                        wd = self.ipdb.watchdog(
                            'RTM_DELRULE', **old_key._asdict()
                        )
                        self.nl.rule('del', **old_key._asdict())
                        wd.wait()
                        with self._direct_state:
                            self['ipdb_scope'] = 'reload'
                    self.nl.rule('add', **transaction)
                transaction.wait_all_targets()
            # rule removal
            if (transaction['ipdb_scope'] in ('shadow', 'remove')) or (
                (transaction['ipdb_scope'] == 'create') and commit_phase == 2
            ):
                if transaction['ipdb_scope'] == 'shadow':
                    with self._direct_state:
                        self['ipdb_scope'] = 'locked'
                # create watchdog
                key = self.make_key(snapshot)
                wd = self.ipdb.watchdog('RTM_DELRULE', **key._asdict())
                self.nl.rule('del', **key._asdict())
                wd.wait()
                if transaction['ipdb_scope'] == 'shadow':
                    with self._direct_state:
                        self['ipdb_scope'] = 'shadow'
            # everything ok
            drop = True

        except Exception as e:
            error = e
            # prepare postmortem
            debug['traceback'] = traceback.format_exc()
            debug['error_stack'] = []
            debug['next_stage'] = None

            if commit_phase == 1:
                try:
                    self.commit(
                        transaction=snapshot,
                        commit_phase=2,
                        commit_mask=commit_mask,
                    )
                except Exception as i_e:
                    debug['next_stage'] = i_e
                    error = RuntimeError()

        if drop and notx:
            self.drop(transaction.uid)

        if error is not None:
            error.debug = debug
            raise error

        return self

    def remove(self):
        self['ipdb_scope'] = 'remove'
        return self

    def shadow(self):
        self['ipdb_scope'] = 'shadow'
        return self


class RulesDict(dict):
    def __init__(self, ipdb):
        self.ipdb = ipdb
        self.lock = threading.Lock()
        self._event_map = {
            'RTM_NEWRULE': self.load_netlink,
            'RTM_DELRULE': self.load_netlink,
        }

    def _register(self):
        for msg in self.ipdb.nl.get_rules(family=AF_INET):
            self.load_netlink(msg)
        for msg in self.ipdb.nl.get_rules(family=AF_INET6):
            self.load_netlink(msg)

    def __getitem__(self, key):
        with self.lock:
            if isinstance(key, RuleKey):
                return super(RulesDict, self).__getitem__(key)
            elif isinstance(key, tuple):
                return super(RulesDict, self).__getitem__(RuleKey(*key))
            elif isinstance(key, int):
                for k in self.keys():
                    if key == k[2]:
                        return super(RulesDict, self).__getitem__(k)
            elif isinstance(key, dict):
                for v in self.values():
                    for k in key:
                        if key[k] != v.get(k, None):
                            break
                    else:
                        return v

    def add(self, spec=None, **kwarg):
        '''
        Create a rule from a dictionary
        '''
        spec = dict(spec or kwarg)
        # action and priority are parts of the key, so
        # they must be specified
        if 'priority' not in spec:
            spec['priority'] = 32000
        if 'table' in spec:
            spec['action'] = FR_ACT_NAMES['FR_ACT_TO_TBL']
        elif 'goto' in spec:
            spec['action'] = FR_ACT_NAMES['FR_ACT_GOTO']
        if 'family' not in spec:
            spec['family'] = AF_INET

        rule = Rule(self.ipdb)
        rule.update(spec)
        # setup the scope
        with rule._direct_state:
            rule['ipdb_scope'] = 'create'
        #
        rule.begin()
        for key, value in spec.items():
            rule[key] = value
        self[rule.make_key(spec)] = rule
        return rule

    def load_netlink(self, msg):
        if not isinstance(msg, fibmsg):
            return

        key = Rule.make_key(msg)

        # RTM_DELRULE
        if msg['event'] == 'RTM_DELRULE':
            try:
                # locate the record
                record = self[key]
                # delete the record
                if record['ipdb_scope'] not in ('locked', 'shadow'):
                    del self[key]
                    with record._direct_state:
                        record['ipdb_scope'] = 'detached'
            except Exception as e:
                # just ignore this failure for now
                log.debug("delrule failed for %s", e)
            return

        # RTM_NEWRULE
        if key not in self:
            self[key] = Rule(self.ipdb)
        self[key].load_netlink(msg)
        return self[key]


spec = [{'name': 'rules', 'class': RulesDict, 'kwarg': {}}]



from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import TC_H_ROOT
from pyroute2.netlink.rtnl.tcmsg.common import get_rate, percent2u32, time2tick

parent = TC_H_ROOT


def get_parameters(kwarg):
    delay = time2tick(kwarg.get('delay', 0))  # in microsecond
    limit = kwarg.get('limit', 1000)  # fifo limit (packets) see netem.c:230
    loss = percent2u32(kwarg.get('loss', 0))  # int percentage
    gap = kwarg.get('gap', 0)
    duplicate = kwarg.get('duplicate', 0)
    jitter = time2tick(kwarg.get('jitter', 0))  # in microsecond

    opts = {
        'delay': delay,
        'limit': limit,
        'loss': loss,
        'gap': gap,
        'duplicate': duplicate,
        'jitter': jitter,
        'attrs': [],
    }

    # correlation (delay, loss, duplicate)
    delay_corr = percent2u32(kwarg.get('delay_corr', 0))
    loss_corr = percent2u32(kwarg.get('loss_corr', 0))
    dup_corr = percent2u32(kwarg.get('dup_corr', 0))
    if delay_corr or loss_corr or dup_corr:
        # delay_corr requires that both jitter and delay are != 0
        if delay_corr and not (delay and jitter):
            raise Exception(
                'delay correlation requires delay' ' and jitter to be set'
            )
        # loss correlation and loss
        if loss_corr and not loss:
            raise Exception('loss correlation requires loss to be set')
        # duplicate correlation and duplicate
        if dup_corr and not duplicate:
            raise Exception(
                'duplicate correlation requires ' 'duplicate to be set'
            )

        opts['attrs'].append(
            [
                'TCA_NETEM_CORR',
                {
                    'delay_corr': delay_corr,
                    'loss_corr': loss_corr,
                    'dup_corr': dup_corr,
                },
            ]
        )

    # reorder (probability, correlation)
    prob_reorder = percent2u32(kwarg.get('prob_reorder', 0))
    corr_reorder = percent2u32(kwarg.get('corr_reorder', 0))
    if prob_reorder != 0:
        # gap defaults to 1 if equal to 0
        if gap == 0:
            opts['gap'] = gap = 1
        opts['attrs'].append(
            [
                'TCA_NETEM_REORDER',
                {'prob_reorder': prob_reorder, 'corr_reorder': corr_reorder},
            ]
        )
    else:
        if gap != 0:
            raise Exception('gap can only be set when prob_reorder is set')
        elif corr_reorder != 0:
            raise Exception(
                'corr_reorder can only be set when ' 'prob_reorder is set'
            )

    # corrupt (probability, correlation)
    prob_corrupt = percent2u32(kwarg.get('prob_corrupt', 0))
    corr_corrupt = percent2u32(kwarg.get('corr_corrupt', 0))
    if prob_corrupt:
        opts['attrs'].append(
            [
                'TCA_NETEM_CORRUPT',
                {'prob_corrupt': prob_corrupt, 'corr_corrupt': corr_corrupt},
            ]
        )
    elif corr_corrupt != 0:
        raise Exception(
            'corr_corrupt can only be set when ' 'prob_corrupt is set'
        )

    # rate (rate, packet_overhead, cell_size, cell_overhead)
    rate = get_rate(kwarg.get('rate', None))
    packet_overhead = kwarg.get('packet_overhead', 0)
    cell_size = kwarg.get('cell_size', 0)
    cell_overhead = kwarg.get('cell_overhead', 0)
    if rate is not None:
        opts['attrs'].append(
            [
                'TCA_NETEM_RATE',
                {
                    'rate': rate,
                    'packet_overhead': packet_overhead,
                    'cell_size': cell_size,
                    'cell_overhead': cell_overhead,
                },
            ]
        )
    elif packet_overhead != 0 or cell_size != 0 or cell_overhead != 0:
        raise Exception(
            'packet_overhead, cell_size and cell_overhead'
            'can only be set when rate is set'
        )

    # TODO
    # delay distribution (dist_size, dist_data)
    return opts


class options(nla):
    nla_map = (
        ('TCA_NETEM_UNSPEC', 'none'),
        ('TCA_NETEM_CORR', 'netem_corr'),
        ('TCA_NETEM_DELAY_DIST', 'none'),
        ('TCA_NETEM_REORDER', 'netem_reorder'),
        ('TCA_NETEM_CORRUPT', 'netem_corrupt'),
        ('TCA_NETEM_LOSS', 'none'),
        ('TCA_NETEM_RATE', 'netem_rate'),
    )

    fields = (
        ('delay', 'I'),
        ('limit', 'I'),
        ('loss', 'I'),
        ('gap', 'I'),
        ('duplicate', 'I'),
        ('jitter', 'I'),
    )

    class netem_corr(nla):
        '''correlation'''

        fields = (('delay_corr', 'I'), ('loss_corr', 'I'), ('dup_corr', 'I'))

    class netem_reorder(nla):
        '''reorder has probability and correlation'''

        fields = (('prob_reorder', 'I'), ('corr_reorder', 'I'))

    class netem_corrupt(nla):
        '''corruption has probability and correlation'''

        fields = (('prob_corrupt', 'I'), ('corr_corrupt', 'I'))

    class netem_rate(nla):
        '''rate'''

        fields = (
            ('rate', 'I'),
            ('packet_overhead', 'i'),
            ('cell_size', 'I'),
            ('cell_overhead', 'i'),
        )



import ctypes
import errno
import fcntl
import socket

from pyroute2.ethtool.common import LinkModeBits

# ethtool/ethtool-copy.h
IFNAMSIZ = 16
SIOCETHTOOL = 0x8946
ETHTOOL_GSET = 0x1
ETHTOOL_GCOALESCE = 0xE
ETHTOOL_SCOALESCE = 0xF
ETHTOOL_GSSET_INFO = 0x37
ETHTOOL_GWOL = 0x00000005

ETHTOOL_GFLAGS = 0x00000025
ETHTOOL_GFEATURES = 0x0000003A
ETHTOOL_SFEATURES = 0x0000003B
ETHTOOL_GLINKSETTINGS = 0x0000004C

ETHTOOL_GSTRINGS = 0x0000001B
ETH_GSTRING_LEN = 32

ETHTOOL_GRXCSUM = 0x00000014
ETHTOOL_SRXCSUM = 0x00000015
ETHTOOL_GTXCSUM = 0x00000016
ETHTOOL_STXCSUM = 0x00000017
ETHTOOL_GSG = 0x00000018
ETHTOOL_SSG = 0x00000019
ETHTOOL_GTSO = 0x0000001E
ETHTOOL_STSO = 0x0000001F
ETHTOOL_GUFO = 0x00000021
ETHTOOL_SUFO = 0x00000022
ETHTOOL_GGSO = 0x00000023
ETHTOOL_SGSO = 0x00000024
ETHTOOL_GGRO = 0x0000002B
ETHTOOL_SGRO = 0x0000002C

SOPASS_MAX = 6

ETH_SS_FEATURES = 4

ETH_FLAG_RXCSUM = 1 << 0
ETH_FLAG_TXCSUM = 1 << 1
ETH_FLAG_SG = 1 << 2
ETH_FLAG_TSO = 1 << 3
ETH_FLAG_UFO = 1 << 4
ETH_FLAG_GSO = 1 << 5
ETH_FLAG_GRO = 1 << 6
ETH_FLAG_TXVLAN = 1 << 7
ETH_FLAG_RXVLAN = 1 << 8
ETH_FLAG_LRO = 1 << 15
ETH_FLAG_NTUPLE = 1 << 27
ETH_FLAG_RXHASH = 1 << 28
ETH_FLAG_EXT_MASK = (
    ETH_FLAG_LRO
    | ETH_FLAG_RXVLAN
    | ETH_FLAG_TXVLAN
    | ETH_FLAG_NTUPLE
    | ETH_FLAG_RXHASH
)

SCHAR_MAX = 127
ETHTOOL_LINK_MODE_MASK_MAX_KERNEL_NU32 = SCHAR_MAX


# Wake-On-Lan options.
WAKE_PHY = 1 << 0
WAKE_UCAST = 1 << 1
WAKE_MCAST = 1 << 2
WAKE_BCAST = 1 << 3
WAKE_ARP = 1 << 4
WAKE_MAGIC = 1 << 5
WAKE_MAGICSECURE = 1 << 6  # only meaningful if WAKE_MAGIC
WAKE_FILTER = 1 << 7
WAKE_NAMES = {
    WAKE_PHY: "phy",
    WAKE_UCAST: "ucast",
    WAKE_MCAST: "mcast",
    WAKE_BCAST: "bcast",
    WAKE_ARP: "arp",
    WAKE_MAGIC: "magic",
    WAKE_MAGICSECURE: "magic_secure",
    WAKE_FILTER: "filter",
}


class EthtoolError(Exception):
    pass


class NotSupportedError(EthtoolError):
    pass


class NoSuchDevice(EthtoolError):
    pass


class DictStruct(ctypes.Structure):
    def __init__(self, *args, **kwargs):
        super(DictStruct, self).__init__(*args, **kwargs)
        self._fields_as_dict = {
            name: [
                lambda k: getattr(self, k),
                lambda k, v: setattr(self, k, v),
            ]
            for name, ct in self._fields_
        }

    def __getitem__(self, key):
        return self._fields_as_dict[key][0](key)

    def __setitem__(self, key, value):
        return self._fields_as_dict[key][1](key, value)

    def __iter__(self):
        return iter(self._fields_as_dict)

    def items(self):
        for k, f in self._fields_as_dict.items():
            getter, _ = f
            yield k, getter(k)

    def keys(self):
        return self._fields_as_dict.keys()

    def __contains__(self, key):
        return key in self._fields_as_dict


class EthtoolWolInfo(DictStruct):
    _fields_ = [
        ("cmd", ctypes.c_uint32),
        ("supported", ctypes.c_uint32),
        ("wolopts", ctypes.c_uint32),
        ("sopass", ctypes.c_uint8 * SOPASS_MAX),
    ]


class EthtoolCmd(DictStruct):
    _pack_ = 1
    _fields_ = [
        ("cmd", ctypes.c_uint32),
        ("supported", ctypes.c_uint32),
        ("advertising", ctypes.c_uint32),
        ("speed", ctypes.c_uint16),
        ("duplex", ctypes.c_uint8),
        ("port", ctypes.c_uint8),
        ("phy_address", ctypes.c_uint8),
        ("transceiver", ctypes.c_uint8),
        ("autoneg", ctypes.c_uint8),
        ("mdio_support", ctypes.c_uint8),
        ("maxtxpkt", ctypes.c_uint32),
        ("maxrxpkt", ctypes.c_uint32),
        ("speed_hi", ctypes.c_uint16),
        ("eth_tp_mdix", ctypes.c_uint8),
        ("reserved2", ctypes.c_uint8),
        ("lp_advertising", ctypes.c_uint32),
        ("reserved", ctypes.c_uint32 * 2),
    ]


class IoctlEthtoolLinkSettings(DictStruct):
    _pack_ = 1
    _fields_ = [
        ("cmd", ctypes.c_uint32),
        ("speed", ctypes.c_uint32),
        ("duplex", ctypes.c_uint8),
        ("port", ctypes.c_uint8),
        ("phy_address", ctypes.c_uint8),
        ("autoneg", ctypes.c_uint8),
        ("mdio_support", ctypes.c_uint8),
        ("eth_tp_mdix", ctypes.c_uint8),
        ("eth_tp_mdix_ctrl", ctypes.c_uint8),
        ("link_mode_masks_nwords", ctypes.c_int8),
        ("transceiver", ctypes.c_uint8),
        ("reserved1", ctypes.c_uint8 * 3),
        ("reserved", ctypes.c_uint32 * 7),
        (
            "link_mode_data",
            ctypes.c_uint32 * (3 * ETHTOOL_LINK_MODE_MASK_MAX_KERNEL_NU32),
        ),
    ]


class EthtoolCoalesce(DictStruct):
    _pack_ = 1
    _fields_ = [
        # ETHTOOL_{G,S}COALESCE
        ("cmd", ctypes.c_uint32),
        # How many usecs to delay an RX interrupt after
        # a packet arrives.  If 0, only rx_max_coalesced_frames
        # is used.
        ("rx_coalesce_usecs", ctypes.c_uint32),
        # How many packets to delay an RX interrupt after
        # a packet arrives.  If 0, only rx_coalesce_usecs is
        # used.  It is illegal to set both usecs and max frames
        # to zero as this would cause RX interrupts to never be
        # generated.
        ("rx_max_coalesced_frames", ctypes.c_uint32),
        # Same as above two parameters, except that these values
        # apply while an IRQ is being serviced by the host.  Not
        # all cards support this feature and the values are ignored
        # in that case.
        ("rx_coalesce_usecs_irq", ctypes.c_uint32),
        ("rx_max_coalesced_frames_irq", ctypes.c_uint32),
        # How many usecs to delay a TX interrupt after
        # a packet is sent.  If 0, only tx_max_coalesced_frames
        # is used.
        ("tx_coalesce_usecs", ctypes.c_uint32),
        # How many packets to delay a TX interrupt after
        # a packet is sent.  If 0, only tx_coalesce_usecs is
        # used.  It is illegal to set both usecs and max frames
        # to zero as this would cause TX interrupts to never be
        # generated.
        ("tx_max_coalesced_frames", ctypes.c_uint32),
        # Same as above two parameters, except that these values
        # apply while an IRQ is being serviced by the host.  Not
        # all cards support this feature and the values are ignored
        # in that case.
        ("tx_coalesce_usecs_irq", ctypes.c_uint32),
        ("tx_max_coalesced_frames_irq", ctypes.c_uint32),
        # How many usecs to delay in-memory statistics
        # block updates.  Some drivers do not have an in-memory
        # statistic block, and in such cases this value is ignored.
        # This value must not be zero.
        ("stats_block_coalesce_usecs", ctypes.c_uint32),
        # Adaptive RX/TX coalescing is an algorithm implemented by
        # some drivers to improve latency under low packet rates and
        # improve throughput under high packet rates.  Some drivers
        # only implement one of RX or TX adaptive coalescing.  Anything
        # not implemented by the driver causes these values to be
        # silently ignored.
        ("use_adaptive_rx_coalesce", ctypes.c_uint32),
        ("use_adaptive_tx_coalesce", ctypes.c_uint32),
        # When the packet rate (measured in packets per second)
        # is below pkt_rate_low, the {rx,tx}_*_low parameters are
        # used.
        ("pkt_rate_low", ctypes.c_uint32),
        ("rx_coalesce_usecs_low", ctypes.c_uint32),
        ("rx_max_coalesced_frames_low", ctypes.c_uint32),
        ("tx_coalesce_usecs_low", ctypes.c_uint32),
        ("tx_max_coalesced_frames_low", ctypes.c_uint32),
        # When the packet rate is below pkt_rate_high but above
        # pkt_rate_low (both measured in packets per second) the
        # normal {rx,tx}_* coalescing parameters are used.
        # When the packet rate is (measured in packets per second)
        # is above pkt_rate_high, the {rx,tx}_*_high parameters are
        # used.
        ("pkt_rate_high", ctypes.c_uint32),
        ("rx_coalesce_usecs_high", ctypes.c_uint32),
        ("rx_max_coalesced_frames_high", ctypes.c_uint32),
        ("tx_coalesce_usecs_high", ctypes.c_uint32),
        ("tx_max_coalesced_frames_high", ctypes.c_uint32),
        # How often to do adaptive coalescing packet rate sampling,
        # measured in seconds.  Must not be zero.
        ("rate_sample_interval", ctypes.c_uint32),
    ]


class EthtoolValue(ctypes.Structure):
    _fields_ = [("cmd", ctypes.c_uint32), ("data", ctypes.c_uint32)]


class EthtoolSsetInfo(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("cmd", ctypes.c_uint32),
        ("reserved", ctypes.c_uint32),
        ("sset_mask", ctypes.c_uint64),
        ("data", ctypes.c_uint32),
    ]


class EthtoolGstrings(ctypes.Structure):
    _fields_ = [
        ("cmd", ctypes.c_uint32),
        ("string_set", ctypes.c_uint32),
        ("len", ctypes.c_uint32),
        # If you have more than 256 features on your NIC
        # they will not be seen by it
        ("strings", ctypes.c_ubyte * ETH_GSTRING_LEN * 256),
    ]


class EthtoolGetFeaturesBlock(ctypes.Structure):
    _fields_ = [
        ("available", ctypes.c_uint32),
        ("requested", ctypes.c_uint32),
        ("active", ctypes.c_uint32),
        ("never_changed", ctypes.c_uint32),
    ]


class EthtoolSetFeaturesBlock(ctypes.Structure):
    _fields_ = [("changed", ctypes.c_uint32), ("active", ctypes.c_uint32)]


def div_round_up(n, d):
    return int(((n) + (d) - 1) / (d))


def feature_bits_to_blocks(n_bits):
    return div_round_up(n_bits, 32)


class EthtoolGfeatures(ctypes.Structure):
    _fields_ = [
        ("cmd", ctypes.c_uint32),
        ("size", ctypes.c_uint32),
        ("features", EthtoolGetFeaturesBlock * feature_bits_to_blocks(256)),
    ]


class EthtoolSfeatures(ctypes.Structure):
    _fields_ = [
        ("cmd", ctypes.c_uint32),
        ("size", ctypes.c_uint32),
        ("features", EthtoolSetFeaturesBlock * feature_bits_to_blocks(256)),
    ]


class FeatureState(ctypes.Structure):
    _fields_ = [("off_flags", ctypes.c_uint32), ("features", EthtoolGfeatures)]


class IfReqData(ctypes.Union):
    _fields_ = [
        ("ifr_data", ctypes.POINTER(EthtoolCmd)),
        ("coalesce", ctypes.POINTER(EthtoolCoalesce)),
        ("value", ctypes.POINTER(EthtoolValue)),
        ("sset_info", ctypes.POINTER(EthtoolSsetInfo)),
        ("gstrings", ctypes.POINTER(EthtoolGstrings)),
        ("gfeatures", ctypes.POINTER(EthtoolGfeatures)),
        ("sfeatures", ctypes.POINTER(EthtoolSfeatures)),
        ("glinksettings", ctypes.POINTER(IoctlEthtoolLinkSettings)),
        ("wolinfo", ctypes.POINTER(EthtoolWolInfo)),
    ]


class IfReq(ctypes.Structure):
    _pack_ = 1
    _anonymous_ = ("u",)
    _fields_ = [("ifr_name", ctypes.c_uint8 * IFNAMSIZ), ("u", IfReqData)]


class EthtoolFeaturesList:
    def __init__(self, cmd, stringsset):
        self._offsets = {}
        self._cmd = cmd
        self._cmd_set = EthtoolSfeatures(cmd=ETHTOOL_SFEATURES, size=cmd.size)
        self._gfeatures = cmd.features
        self._sfeatures = self._cmd_set.features

        feature_i = 0
        for i, name in enumerate(stringsset):
            feature_i = i // 32
            flag_bit = 1 << (i % 32)
            self._offsets[name] = (feature_i, flag_bit)

        while feature_i:
            feature_i -= 1
            self._sfeatures[feature_i].active = self._gfeatures[
                feature_i
            ].active
            self._sfeatures[feature_i].changed = 0

    def is_available(self, name):
        feature_i, flag_bit = self._offsets[name]
        return self._gfeatures[feature_i].available & flag_bit != 0

    def is_active(self, name):
        feature_i, flag_bit = self._offsets[name]
        return self._gfeatures[feature_i].active & flag_bit != 0

    def is_requested(self, name):
        feature_i, flag_bit = self._offsets[name]
        return self._gfeatures[feature_i].requested & flag_bit != 0

    def is_never_changed(self, name):
        feature_i, flag_bit = self._offsets[name]
        return self._gfeatures[feature_i].never_changed & flag_bit != 0

    def __iter__(self):
        for name in self._offsets:
            feature_i, flag_bit = self._offsets[name]
            yield (
                name,
                self.get_value(name),
                self.is_available(name),
                feature_i,
                flag_bit,
            )

    def keys(self):
        return self._offsets.keys()

    def __contains__(self, name):
        return name in self._offsets

    def __getitem__(self, key):
        return self.get_value(key)

    def __setitem__(self, key, value):
        return self.set_value(key, value)

    def get_value(self, name):
        return self.is_active(name)

    def set_value(self, name, value):
        if value not in (1, 0, True, False):
            raise ValueError("Need a boolean value")

        feature_i, flag_bit = self._offsets[name]
        if value:
            self._gfeatures[feature_i].active |= flag_bit
            self._sfeatures[feature_i].active |= flag_bit
        else:
            # active is ctypes.c_uint32
            self._gfeatures[feature_i].active &= flag_bit ^ 0xFFFFFFFF
            self._sfeatures[feature_i].active &= flag_bit ^ 0xFFFFFFFF
        self._sfeatures[feature_i].changed |= flag_bit


class IoctlEthtool:
    def __init__(self, ifname=None):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.ifname = None
        self.ifreq = None

        if ifname is not None:
            self.change_ifname(ifname)

    def close(self):
        self.sock.close()

    def change_ifname(self, ifname):
        self.ifname = bytearray(ifname, 'utf-8')
        self.ifname.extend(b"\0" * (IFNAMSIZ - len(self.ifname)))
        self.ifreq = IfReq()
        self.ifreq.ifr_name = (ctypes.c_uint8 * IFNAMSIZ)(*self.ifname)

    def ioctl(self):
        try:
            if fcntl.ioctl(self.sock, SIOCETHTOOL, self.ifreq):
                raise NotSupportedError()
        except OSError as e:
            if e.errno == errno.ENOTSUP:
                raise NotSupportedError(self.ifname.decode("utf-8"))
            elif e.errno == errno.ENODEV:
                raise NoSuchDevice(self.ifname.decode("utf-8"))
            raise

    def get_stringset(
        self, set_id=ETH_SS_FEATURES, drvinfo_offset=0, null_terminate=1
    ):
        sset_info = EthtoolSsetInfo(
            cmd=ETHTOOL_GSSET_INFO, reserved=0, sset_mask=1 << set_id
        )
        self.ifreq.sset_info = ctypes.pointer(sset_info)
        fcntl.ioctl(self.sock, SIOCETHTOOL, self.ifreq)
        if sset_info.sset_mask:
            length = sset_info.data
        else:
            length = 0

        strings_found = []
        gstrings = EthtoolGstrings(
            cmd=ETHTOOL_GSTRINGS, string_set=set_id, len=length
        )
        self.ifreq.gstrings = ctypes.pointer(gstrings)
        self.ioctl()

        for i in range(length):
            buf = ''
            for j in range(ETH_GSTRING_LEN):
                code = gstrings.strings[i][j]
                if code == 0:
                    break
                buf += chr(code)
            strings_found.append(buf)
        return strings_found

    def get_features(self):
        stringsset = self.get_stringset()
        cmd = EthtoolGfeatures()
        cmd.cmd = ETHTOOL_GFEATURES
        cmd.size = feature_bits_to_blocks(len(stringsset))
        self.ifreq.gfeatures = ctypes.pointer(cmd)
        self.ioctl()
        return EthtoolFeaturesList(cmd, stringsset)

    def set_features(self, features):
        self.ifreq.sfeatures = ctypes.pointer(features._cmd_set)
        return self.ioctl()

    def get_cmd(self):
        cmd = EthtoolCmd(cmd=ETHTOOL_GSET)
        self.ifreq.ifr_data = ctypes.pointer(cmd)
        self.ioctl()
        return cmd

    @staticmethod
    def get_link_mode_bits(map_bits):
        for bit in LinkModeBits:
            map_i = bit.bit_index // 32
            map_bit = bit.bit_index % 32
            if map_i >= len(map_bits):
                continue

            if map_bits[map_i] & (1 << map_bit):
                yield bit

    @staticmethod
    def get_link_mode_masks(ecmd):
        map_supported = []
        map_advertising = []
        map_lp_advertising = []
        i = 0
        while i != ecmd.link_mode_masks_nwords:
            map_supported.append(ecmd.link_mode_data[i])
            i += 1
        while i != ecmd.link_mode_masks_nwords * 2:
            map_advertising.append(ecmd.link_mode_data[i])
            i += 1
        while i != ecmd.link_mode_masks_nwords * 3:
            map_lp_advertising.append(ecmd.link_mode_data[i])
            i += 1

        return (map_supported, map_advertising, map_lp_advertising)

    def get_link_settings(self):
        ecmd = IoctlEthtoolLinkSettings()
        ecmd.cmd = ETHTOOL_GLINKSETTINGS
        self.ifreq.glinksettings = ctypes.pointer(ecmd)

        # Handshake with kernel to determine number of words for link
        # mode bitmaps. When requested number of bitmap words is not
        # the one expected by kernel, the latter returns the integer
        # opposite of what it is expecting. We request length 0 below
        # (aka. invalid bitmap length) to get this info.
        self.ioctl()

        # see above: we expect a strictly negative value from kernel.
        if (
            ecmd.link_mode_masks_nwords >= 0
            or ecmd.cmd != ETHTOOL_GLINKSETTINGS
        ):
            raise NotSupportedError()

        # got the real ecmd.req.link_mode_masks_nwords,
        # now send the real request
        ecmd.link_mode_masks_nwords = -ecmd.link_mode_masks_nwords
        self.ioctl()

        if (
            ecmd.link_mode_masks_nwords <= 0
            or ecmd.cmd != ETHTOOL_GLINKSETTINGS
        ):
            raise NotSupportedError()

        return ecmd

    def get_coalesce(self):
        cmd = EthtoolCoalesce(cmd=ETHTOOL_GCOALESCE)
        self.ifreq.coalesce = ctypes.pointer(cmd)
        self.ioctl()
        return cmd

    def set_coalesce(self, coalesce):
        coalesce.cmd = ETHTOOL_SCOALESCE
        self.ifreq.coalesce = ctypes.pointer(coalesce)
        self.ioctl()
        return

    def get_wol(self):
        cmd = EthtoolWolInfo(cmd=ETHTOOL_GWOL)
        self.ifreq.wolinfo = ctypes.pointer(cmd)
        self.ioctl()
        return cmd



import struct

from pyroute2.netlink import nlmsg_base, nlmsg_decoder_generic


class inotify_msg(nlmsg_base, nlmsg_decoder_generic):
    fields = (
        ('wd', 'i'),
        ('mask', 'I'),
        ('cookie', 'I'),
        ('name_length', 'I'),
    )

    def decode(self):
        super(inotify_msg, self).decode()
        (name,) = struct.unpack_from(
            '%is' % self['name_length'], self.data, self.offset + 16
        )
        self['name'] = name.decode('utf-8').strip('\0')
        self.length = self['name_length'] + 16



'''
'''
from pyroute2.netlink.nfnetlink import nfgen_msg
from pyroute2.netlink.nfnetlink.nftsocket import (
    DATA_TYPE_ID_TO_NAME,
    DATA_TYPE_NAME_TO_INFO,
    NFT_MSG_DELCHAIN,
    NFT_MSG_DELRULE,
    NFT_MSG_DELSET,
    NFT_MSG_DELSETELEM,
    NFT_MSG_DELTABLE,
    NFT_MSG_GETCHAIN,
    NFT_MSG_GETRULE,
    NFT_MSG_GETSET,
    NFT_MSG_GETSETELEM,
    NFT_MSG_GETTABLE,
    NFT_MSG_NEWCHAIN,
    NFT_MSG_NEWRULE,
    NFT_MSG_NEWSET,
    NFT_MSG_NEWSETELEM,
    NFT_MSG_NEWTABLE,
    NFTSocket,
    nft_chain_msg,
    nft_rule_msg,
    nft_set_elem_list_msg,
    nft_set_msg,
    nft_table_msg,
)


class NFTSet:
    __slots__ = ('table', 'name', 'key_type', 'timeout', 'counter', 'comment')

    def __init__(self, table, name, **kwargs):
        self.table = table
        self.name = name

        for attrname in self.__slots__:
            if attrname in kwargs:
                setattr(self, attrname, kwargs[attrname])
            elif attrname not in ("table", "name"):
                setattr(self, attrname, None)

    def as_netlink(self):
        attrs = {"NFTA_SET_TABLE": self.table, "NFTA_SET_NAME": self.name}
        set_flags = set()

        if self.key_type is not None:
            key_type, key_len, _ = DATA_TYPE_NAME_TO_INFO.get(self.key_type)
            attrs["NFTA_SET_KEY_TYPE"] = key_type
            attrs["NFTA_SET_KEY_LEN"] = key_len

        if self.timeout is not None:
            set_flags.add("NFT_SET_TIMEOUT")
            attrs["NFTA_SET_TIMEOUT"] = self.timeout

        if self.counter is True:
            attrs["NFTA_SET_EXPR"] = {'attrs': [('NFTA_EXPR_NAME', 'counter')]}

        if self.comment is not None:
            attrs["NFTA_SET_USERDATA"] = [
                ("NFTNL_UDATA_SET_COMMENT", self.comment)
            ]

        # ID is used for bulk create, but not implemented
        attrs['NFTA_SET_ID'] = 1
        attrs["NFTA_SET_FLAGS"] = set_flags
        return attrs

    @classmethod
    def from_netlink(cls, msg):
        data_type_name = DATA_TYPE_ID_TO_NAME.get(
            msg.get_attr("NFTA_SET_KEY_TYPE"),
            msg.get_attr("NFTA_SET_KEY_TYPE"),  # fallback to raw value
        )

        counter = False
        expr = msg.get_attr('NFTA_SET_EXPR')
        if expr:
            expr = expr.get_attrs('NFTA_EXPR_NAME')
            if expr and "counter" in expr:
                counter = True

        comment = None
        udata = msg.get_attr("NFTA_SET_USERDATA")
        if udata:
            for key, value in udata:
                if key == "NFTNL_UDATA_SET_COMMENT":
                    comment = value
                    break

        return cls(
            table=msg.get_attr('NFTA_SET_TABLE'),
            name=msg.get_attr('NFTA_SET_NAME'),
            key_type=data_type_name,
            timeout=msg.get_attr('NFTA_SET_TIMEOUT'),
            counter=counter,
            comment=comment,
        )

    @classmethod
    def from_dict(cls, d):
        return cls(
            **{
                name: value
                for name, value in d.items()
                if name in cls.__slots__
            }
        )

    def as_dict(self):
        return {name: getattr(self, name) for name in self.__slots__}

    def __repr__(self):
        return str(self.as_dict())


class NFTSetElem:
    __slots__ = (
        'value',
        'timeout',
        'expiration',
        'counter_bytes',
        'counter_packets',
        'comment',
    )

    def __init__(self, value, **kwargs):
        self.value = value
        for name in self.__slots__:
            if name in kwargs:
                setattr(self, name, kwargs[name])
            elif name != "value":
                setattr(self, name, None)

    @classmethod
    def from_netlink(cls, msg, modifier):
        value = msg.get_attr('NFTA_SET_ELEM_KEY').get_attr("NFTA_DATA_VALUE")
        if modifier is not None:
            # Need to find a better way
            modifier.data = value
            modifier.length = 4 + len(modifier.data)
            modifier.decode()
            value = modifier.value

        kwarg = {
            "expiration": msg.get_attr('NFTA_SET_ELEM_EXPIRATION'),
            "timeout": msg.get_attr('NFTA_SET_ELEM_TIMEOUT'),
        }

        elem_expr = msg.get_attr('NFTA_SET_ELEM_EXPR')
        if elem_expr:
            if elem_expr.get_attr('NFTA_EXPR_NAME') == "counter":
                elem_expr = elem_expr.get_attr("NFTA_EXPR_DATA")
                kwarg.update(
                    {
                        "counter_bytes": elem_expr.get_attr(
                            "NFTA_COUNTER_BYTES"
                        ),
                        "counter_packets": elem_expr.get_attr(
                            "NFTA_COUNTER_PACKETS"
                        ),
                    }
                )

        udata = msg.get_attr('NFTA_SET_ELEM_USERDATA')
        if udata:
            for type_name, data in udata:
                if type_name == "NFTNL_UDATA_SET_ELEM_COMMENT":
                    kwarg["comment"] = data

        return cls(value=value, **kwarg)

    def as_netlink(self, modifier):
        if modifier is not None:
            modifier.value = self.value
            modifier.encode()
            value = modifier["value"]
        else:
            value = self.value

        attrs = [
            ['NFTA_SET_ELEM_KEY', {'attrs': [('NFTA_DATA_VALUE', value)]}]
        ]

        if self.timeout is not None:
            attrs.append(['NFTA_SET_ELEM_TIMEOUT', self.timeout])

        if self.expiration is not None:
            attrs.append(['NFTA_SET_ELEM_EXPIRATION', self.expiration])

        if self.comment is not None:
            attrs.append(
                [
                    'NFTA_SET_ELEM_USERDATA',
                    [("NFTNL_UDATA_SET_ELEM_COMMENT", self.comment)],
                ]
            )

        return {'attrs': attrs}

    @classmethod
    def from_dict(cls, d):
        return cls(
            **{
                name: value
                for name, value in d.items()
                if name in cls.__slots__
            }
        )

    def as_dict(self):
        return {name: getattr(self, name) for name in self.__slots__}

    def __repr__(self):
        return str(self.as_dict())


class NFTables(NFTSocket):
    # TODO: documentation
    # TODO: tests
    # TODO: dump()/load() with support for json and xml

    def get_tables(self):
        return self.request_get(nfgen_msg(), NFT_MSG_GETTABLE)

    def get_chains(self):
        return self.request_get(nfgen_msg(), NFT_MSG_GETCHAIN)

    def get_rules(self):
        return self.request_get(nfgen_msg(), NFT_MSG_GETRULE)

    def get_sets(self):
        return self.request_get(nfgen_msg(), NFT_MSG_GETSET)

    #
    # The nft API is in the prototype stage and may be
    # changed until the release. The planned release for
    # the API is 0.5.2
    #

    def table(self, cmd, **kwarg):
        '''
        Example::

            nft.table('add', name='test0')
        '''
        commands = {
            'add': NFT_MSG_NEWTABLE,
            'create': NFT_MSG_NEWTABLE,
            'del': NFT_MSG_DELTABLE,
            'get': NFT_MSG_GETTABLE,
        }
        return self._command(nft_table_msg, commands, cmd, kwarg)

    def chain(self, cmd, **kwarg):
        '''
        Example::

            #
            # default policy 'drop' for input
            #
            nft.chain('add',
                      table='test0',
                      name='test_chain0',
                      hook='input',
                      type='filter',
                      policy=0)
        '''
        commands = {
            'add': NFT_MSG_NEWCHAIN,
            'create': NFT_MSG_NEWCHAIN,
            'del': NFT_MSG_DELCHAIN,
            'get': NFT_MSG_GETCHAIN,
        }
        # TODO: What about 'ingress' (netdev family)?
        hooks = {
            'prerouting': 0,
            'input': 1,
            'forward': 2,
            'output': 3,
            'postrouting': 4,
        }
        if 'hook' in kwarg:
            kwarg['hook'] = {
                'attrs': [
                    ['NFTA_HOOK_HOOKNUM', hooks[kwarg['hook']]],
                    ['NFTA_HOOK_PRIORITY', kwarg.pop('priority', 0)],
                ]
            }
        if 'type' not in kwarg:
            kwarg['type'] = 'filter'
        return self._command(nft_chain_msg, commands, cmd, kwarg)

    def rule(self, cmd, **kwarg):
        '''
        Example::

            from pyroute2.nftables.expressions import ipv4addr, verdict
            #
            # allow all traffic from 192.168.0.0/24
            #
            nft.rule('add',
                     table='test0',
                     chain='test_chain0',
                     expressions=(ipv4addr(src='192.168.0.0/24'),
                                  verdict(code=1)))
        '''
        commands = {
            'add': NFT_MSG_NEWRULE,
            'create': NFT_MSG_NEWRULE,
            'insert': NFT_MSG_NEWRULE,
            'replace': NFT_MSG_NEWRULE,
            'del': NFT_MSG_DELRULE,
            'get': NFT_MSG_GETRULE,
        }

        if 'expressions' in kwarg:
            expressions = []
            for exp in kwarg['expressions']:
                expressions.extend(exp)
            kwarg['expressions'] = expressions
        return self._command(nft_rule_msg, commands, cmd, kwarg)

    def sets(self, cmd, **kwarg):
        '''
        Example::
            nft.sets("add", table="filter", name="test0", key_type="ipv4_addr",
                     timeout=10000, counter=True,
                     comment="my comment max 252 bytes")
            nft.sets("get", table="filter", name="test0")
            nft.sets("del", table="filter", name="test0")
            my_set = nft.sets("add", set=NFTSet(table="filter", name="test1",
                              key_type="ipv4_addr")
            nft.sets("del", set=my_set)
        '''
        commands = {
            'add': NFT_MSG_NEWSET,
            'get': NFT_MSG_GETSET,
            'del': NFT_MSG_DELSET,
        }

        if "set" in kwarg:
            nft_set = kwarg.pop("set")
        else:
            nft_set = NFTSet(**kwarg)
        kwarg = nft_set.as_netlink()
        msg = self._command(nft_set_msg, commands, cmd, kwarg)
        if cmd == "get":
            return NFTSet.from_netlink(msg)
        return nft_set

    def set_elems(self, cmd, **kwarg):
        '''
        Example::
            nft.set_elems("add", table="filter", set="test0",
                          elements={"10.2.3.4", "10.4.3.2"})
            nft.set_elems("add", set=NFTSet(table="filter", name="test0"),
                          elements=[{"value": "10.2.3.4", "timeout": 10000}])
            nft.set_elems("add", table="filter", set="test0",
                          elements=[NFTSetElem(value="10.2.3.4",
                                               timeout=10000,
                                               comment="hello world")])
            nft.set_elems("get", table="filter", set="test0")
            nft.set_elems("del", table="filter", set="test0",
                          elements=["10.2.3.4"])
        '''
        commands = {
            'add': NFT_MSG_NEWSETELEM,
            'get': NFT_MSG_GETSETELEM,
            'del': NFT_MSG_DELSETELEM,
        }
        if isinstance(kwarg["set"], NFTSet):
            nft_set = kwarg.pop("set")
            kwarg["table"] = nft_set.table
            kwarg["set"] = nft_set.name
        else:
            nft_set = self.sets("get", table=kwarg["table"], name=kwarg["set"])

        found = DATA_TYPE_NAME_TO_INFO.get(nft_set.key_type)
        if found:
            _, _, modifier = found
            modifier = modifier()
            modifier.header = None
        else:
            modifier = None

        if cmd == "get":
            msg = nft_set_elem_list_msg()
            msg['attrs'] = [
                ["NFTA_SET_ELEM_LIST_TABLE", kwarg["table"]],
                ["NFTA_SET_ELEM_LIST_SET", kwarg["set"]],
            ]
            msg = self.request_get(msg, NFT_MSG_GETSETELEM)[0]
            elements = set()
            for elem in msg.get_attr('NFTA_SET_ELEM_LIST_ELEMENTS'):
                elements.add(NFTSetElem.from_netlink(elem, modifier))
            return elements

        elements = []
        for elem in kwarg.pop("elements"):
            if isinstance(elem, dict):
                elem = NFTSetElem.from_dict(elem)
            elif not isinstance(elem, NFTSetElem):
                elem = NFTSetElem(value=elem)
            elements.append(elem.as_netlink(modifier))
        kwarg["elements"] = elements
        return self._command(nft_set_elem_list_msg, commands, cmd, kwarg)



from pyroute2.netlink import nla


class tun(nla):
    prefix = 'IFLA_'
    nla_map = (
        ('IFLA_TUN_UNSPEC', 'none'),
        ('IFLA_TUN_OWNER', 'uint32'),
        ('IFLA_TUN_GROUP', 'uint32'),
        ('IFLA_TUN_TYPE', 'uint8'),
        ('IFLA_TUN_PI', 'uint8'),
        ('IFLA_TUN_VNET_HDR', 'uint8'),
        ('IFLA_TUN_PERSIST', 'uint8'),
        ('IFLA_TUN_MULTI_QUEUE', 'uint8'),
        ('IFLA_TUN_NUM_QUEUES', 'uint32'),
        ('IFLA_TUN_NUM_DISABLED_QUEUES', 'uint32'),
    )



'''
ipset support.

This module is tested with hash:ip, hash:net, list:set and several
other ipset structures (like hash:net,iface). There is no guarantee
that this module is working with all available ipset modules.

It supports almost all kernel commands (create, destroy, flush,
rename, swap, test...)
'''
import errno
import socket

from pyroute2.common import basestring
from pyroute2.netlink import (
    NETLINK_NETFILTER,
    NLM_F_ACK,
    NLM_F_DUMP,
    NLM_F_EXCL,
    NLM_F_REQUEST,
    NLMSG_ERROR,
)
from pyroute2.netlink.exceptions import IPSetError, NetlinkError
from pyroute2.netlink.nfnetlink import NFNL_SUBSYS_IPSET
from pyroute2.netlink.nfnetlink.ipset import (
    IPSET_CMD_ADD,
    IPSET_CMD_CREATE,
    IPSET_CMD_DEL,
    IPSET_CMD_DESTROY,
    IPSET_CMD_FLUSH,
    IPSET_CMD_GET_BYINDEX,
    IPSET_CMD_GET_BYNAME,
    IPSET_CMD_HEADER,
    IPSET_CMD_LIST,
    IPSET_CMD_PROTOCOL,
    IPSET_CMD_RENAME,
    IPSET_CMD_SWAP,
    IPSET_CMD_TEST,
    IPSET_CMD_TYPE,
    IPSET_ERR_BUSY,
    IPSET_ERR_COMMENT,
    IPSET_ERR_COUNTER,
    IPSET_ERR_EXIST,
    IPSET_ERR_EXIST_SETNAME2,
    IPSET_ERR_FIND_TYPE,
    IPSET_ERR_INVALID_CIDR,
    IPSET_ERR_INVALID_FAMILY,
    IPSET_ERR_INVALID_MARKMASK,
    IPSET_ERR_INVALID_NETMASK,
    IPSET_ERR_IPADDR_IPV4,
    IPSET_ERR_IPADDR_IPV6,
    IPSET_ERR_MAX_SETS,
    IPSET_ERR_PROTOCOL,
    IPSET_ERR_REFERENCED,
    IPSET_ERR_SKBINFO,
    IPSET_ERR_TIMEOUT,
    IPSET_ERR_TYPE_MISMATCH,
    IPSET_FLAG_IFACE_WILDCARD,
    IPSET_FLAG_PHYSDEV,
    IPSET_FLAG_WITH_COMMENT,
    IPSET_FLAG_WITH_COUNTERS,
    IPSET_FLAG_WITH_FORCEADD,
    IPSET_FLAG_WITH_SKBINFO,
    ipset_msg,
)
from pyroute2.netlink.nlsocket import NetlinkSocket


def _nlmsg_error(msg):
    return msg['header']['type'] == NLMSG_ERROR


class PortRange(object):
    """A simple container for port range with optional protocol

    Note that optional protocol parameter is not supported by all
    kernel ipset modules using ports. On the other hand, it's sometimes
    mandatory to set it (like for hash:net,port ipsets)

    Example::

        udp_proto = socket.getprotobyname("udp")
        port_range = PortRange(1000, 2000, protocol=udp_proto)
        ipset.create("foo", stype="hash:net,port")
        ipset.add("foo", ("192.0.2.0/24", port_range), etype="net,port")
        ipset.test("foo", ("192.0.2.0/24", port_range), etype="net,port")
    """

    def __init__(self, begin, end, protocol=None):
        self.begin = begin
        self.end = end
        self.protocol = protocol


class PortEntry(object):
    """A simple container for port entry with optional protocol"""

    def __init__(self, port, protocol=None):
        self.port = port
        self.protocol = protocol


class IPSet(NetlinkSocket):
    '''
    NFNetlink socket (family=NETLINK_NETFILTER).

    Implements API to the ipset functionality.
    '''

    policy = {
        IPSET_CMD_PROTOCOL: ipset_msg,
        IPSET_CMD_LIST: ipset_msg,
        IPSET_CMD_TYPE: ipset_msg,
        IPSET_CMD_HEADER: ipset_msg,
        IPSET_CMD_GET_BYNAME: ipset_msg,
        IPSET_CMD_GET_BYINDEX: ipset_msg,
    }

    attr_map = {
        'iface': 'IPSET_ATTR_IFACE',
        'mark': 'IPSET_ATTR_MARK',
        'set': 'IPSET_ATTR_NAME',
        'mac': 'IPSET_ATTR_ETHER',
        'port': 'IPSET_ATTR_PORT',
        ('ip_from', 1): 'IPSET_ATTR_IP_FROM',
        ('ip_from', 2): 'IPSET_ATTR_IP2',
        ('cidr', 1): 'IPSET_ATTR_CIDR',
        ('cidr', 2): 'IPSET_ATTR_CIDR2',
        ('ip_to', 1): 'IPSET_ATTR_IP_TO',
        ('ip_to', 2): 'IPSET_ATTR_IP2_TO',
    }

    def __init__(self, version=None, attr_revision=None, nfgen_family=2):
        super(IPSet, self).__init__(family=NETLINK_NETFILTER)
        policy = dict(
            [
                (x | (NFNL_SUBSYS_IPSET << 8), y)
                for (x, y) in self.policy.items()
            ]
        )
        self.register_policy(policy)
        self._nfgen_family = nfgen_family
        if version is None:
            msg = self.get_proto_version()
            version = msg[0].get_attr('IPSET_ATTR_PROTOCOL')
        self._proto_version = version
        self._attr_revision = attr_revision

    def request(
        self,
        msg,
        msg_type,
        msg_flags=NLM_F_REQUEST | NLM_F_DUMP,
        terminate=None,
    ):
        msg['nfgen_family'] = self._nfgen_family
        try:
            return tuple(
                self.nlm_request(
                    msg,
                    msg_type | (NFNL_SUBSYS_IPSET << 8),
                    msg_flags,
                    terminate=terminate,
                )
            )
        except NetlinkError as err:
            raise _IPSetError(err.code, cmd=msg_type)

    def headers(self, name, **kwargs):
        '''
        Get headers of the named ipset. It can be used to test if one ipset
        exists, since it returns a no such file or directory.
        '''
        return self._list_or_headers(IPSET_CMD_HEADER, name=name, **kwargs)

    def get_proto_version(self, version=6):
        '''
        Get supported protocol version by kernel.

        version parameter allow to set mandatory (but unused?)
        IPSET_ATTR_PROTOCOL netlink attribute in the request.
        '''
        msg = ipset_msg()
        msg['attrs'] = [['IPSET_ATTR_PROTOCOL', version]]
        return self.request(msg, IPSET_CMD_PROTOCOL)

    def list(self, *argv, **kwargs):
        '''
        List installed ipsets. If `name` is provided, list
        the named ipset or return an empty list.

        Be warned: netlink does not return an error if given name does not
        exit, you will receive an empty list.
        '''
        if argv:
            kwargs['name'] = argv[0]
        return self._list_or_headers(IPSET_CMD_LIST, **kwargs)

    def _list_or_headers(self, cmd, name=None, flags=None):
        msg = ipset_msg()
        msg['attrs'] = [['IPSET_ATTR_PROTOCOL', self._proto_version]]
        if name is not None:
            msg['attrs'].append(['IPSET_ATTR_SETNAME', name])
        if flags is not None:
            msg['attrs'].append(['IPSET_ATTR_FLAGS', flags])
        return self.request(msg, cmd)

    def destroy(self, name=None):
        '''
        Destroy one (when name is set) or all ipset (when name is None)
        '''
        msg = ipset_msg()
        msg['attrs'] = [['IPSET_ATTR_PROTOCOL', self._proto_version]]
        if name is not None:
            msg['attrs'].append(['IPSET_ATTR_SETNAME', name])
        return self.request(
            msg,
            IPSET_CMD_DESTROY,
            msg_flags=NLM_F_REQUEST | NLM_F_ACK | NLM_F_EXCL,
            terminate=_nlmsg_error,
        )

    def create(
        self,
        name,
        stype='hash:ip',
        family=socket.AF_INET,
        exclusive=True,
        counters=False,
        comment=False,
        maxelem=None,
        forceadd=False,
        hashsize=None,
        timeout=None,
        bitmap_ports_range=None,
        size=None,
        skbinfo=False,
    ):
        '''
        Create an ipset `name` of type `stype`, by default
        `hash:ip`.

        Common ipset options are supported:

        * exclusive -- if set, raise an error if the ipset exists
        * counters -- enable data/packets counters
        * comment -- enable comments capability
        * maxelem -- max size of the ipset
        * forceadd -- you should refer to the ipset manpage
        * hashsize -- size of the hashtable (if any)
        * timeout -- enable and set a default value for entries (if not None)
        * bitmap_ports_range -- set the specified inclusive portrange for
                                the bitmap ipset structure (0, 65536)
        * size -- Size of the list:set, the default is 8
        * skbinfo -- enable skbinfo capability
        '''
        excl_flag = NLM_F_EXCL if exclusive else 0
        msg = ipset_msg()
        cadt_flags = 0
        if counters:
            cadt_flags |= IPSET_FLAG_WITH_COUNTERS
        if comment:
            cadt_flags |= IPSET_FLAG_WITH_COMMENT
        if forceadd:
            cadt_flags |= IPSET_FLAG_WITH_FORCEADD
        if skbinfo:
            cadt_flags |= IPSET_FLAG_WITH_SKBINFO

        if stype == 'bitmap:port' and bitmap_ports_range is None:
            raise ValueError('Missing value bitmap_ports_range')

        data = {'attrs': []}
        if cadt_flags:
            data['attrs'] += [['IPSET_ATTR_CADT_FLAGS', cadt_flags]]
        if maxelem is not None:
            data['attrs'] += [['IPSET_ATTR_MAXELEM', maxelem]]
        if hashsize is not None:
            data['attrs'] += [["IPSET_ATTR_HASHSIZE", hashsize]]
        elif size is not None and stype == 'list:set':
            data['attrs'] += [['IPSET_ATTR_SIZE', size]]
        if timeout is not None:
            data['attrs'] += [["IPSET_ATTR_TIMEOUT", timeout]]
        if bitmap_ports_range is not None and stype == 'bitmap:port':
            # Set the bitmap range A bitmap type of set
            # can store up to 65536 entries
            if isinstance(bitmap_ports_range, PortRange):
                data['attrs'] += [
                    ['IPSET_ATTR_PORT_FROM', bitmap_ports_range.begin]
                ]
                data['attrs'] += [
                    ['IPSET_ATTR_PORT_TO', bitmap_ports_range.end]
                ]
            else:
                data['attrs'] += [
                    ['IPSET_ATTR_PORT_FROM', bitmap_ports_range[0]]
                ]
                data['attrs'] += [
                    ['IPSET_ATTR_PORT_TO', bitmap_ports_range[1]]
                ]

        if self._attr_revision is None:
            # Get the last revision supported by kernel
            revision = self.get_supported_revisions(stype)[1]
        else:
            revision = self._attr_revision
        msg['attrs'] = [
            ['IPSET_ATTR_PROTOCOL', self._proto_version],
            ['IPSET_ATTR_SETNAME', name],
            ['IPSET_ATTR_TYPENAME', stype],
            ['IPSET_ATTR_FAMILY', family],
            ['IPSET_ATTR_REVISION', revision],
            ["IPSET_ATTR_DATA", data],
        ]

        return self.request(
            msg,
            IPSET_CMD_CREATE,
            msg_flags=NLM_F_REQUEST | NLM_F_ACK | excl_flag,
            terminate=_nlmsg_error,
        )

    @staticmethod
    def _family_to_version(family):
        if family is not None:
            if family == socket.AF_INET:
                return 'IPSET_ATTR_IPADDR_IPV4'
            elif family == socket.AF_INET6:
                return 'IPSET_ATTR_IPADDR_IPV6'
            elif family == socket.AF_UNSPEC:
                return None
            raise TypeError('unknown family')

    def _entry_to_data_attrs(self, entry, etype, ip_version):
        attrs = []
        ip_count = 0

        if etype == 'set':
            attrs += [['IPSET_ATTR_NAME', entry]]
            return attrs

        # We support string (for one element, and for users calling this
        # function like a command line), and tuple/list
        if isinstance(entry, basestring):
            entry = entry.split(',')
        if isinstance(entry, (int, PortRange, PortEntry)):
            entry = [entry]

        for e, t in zip(entry, etype.split(',')):
            if t in ('ip', 'net'):
                ip_count += 1
                if t == 'net':
                    if '/' in e:
                        e, cidr = e.split('/')
                        attrs += [
                            [self.attr_map[('cidr', ip_count)], int(cidr)]
                        ]
                    elif '-' in e:
                        e, to = e.split('-')
                        attrs += [
                            [
                                self.attr_map[('ip_to', ip_count)],
                                {'attrs': [[ip_version, to]]},
                            ]
                        ]
                attrs += [
                    [
                        self.attr_map[('ip_from', ip_count)],
                        {'attrs': [[ip_version, e]]},
                    ]
                ]
            elif t == "port":
                if isinstance(e, PortRange):
                    attrs += [['IPSET_ATTR_PORT_FROM', e.begin]]
                    attrs += [['IPSET_ATTR_PORT_TO', e.end]]
                    if e.protocol is not None:
                        attrs += [['IPSET_ATTR_PROTO', e.protocol]]
                elif isinstance(e, PortEntry):
                    attrs += [['IPSET_ATTR_PORT', e.port]]
                    if e.protocol is not None:
                        attrs += [['IPSET_ATTR_PROTO', e.protocol]]
                else:
                    attrs += [[self.attr_map[t], e]]
            else:
                attrs += [[self.attr_map[t], e]]

        return attrs

    def _add_delete_test(
        self,
        name,
        entry,
        family,
        cmd,
        exclusive,
        comment=None,
        timeout=None,
        etype="ip",
        packets=None,
        bytes=None,
        skbmark=None,
        skbprio=None,
        skbqueue=None,
        wildcard=False,
        physdev=False,
    ):
        excl_flag = NLM_F_EXCL if exclusive else 0
        adt_flags = 0
        if wildcard:
            adt_flags |= IPSET_FLAG_IFACE_WILDCARD
        if physdev:
            adt_flags |= IPSET_FLAG_PHYSDEV

        ip_version = self._family_to_version(family)
        data_attrs = self._entry_to_data_attrs(entry, etype, ip_version)
        if comment is not None:
            data_attrs += [
                ["IPSET_ATTR_COMMENT", comment],
                ["IPSET_ATTR_CADT_LINENO", 0],
            ]
        if timeout is not None:
            data_attrs += [["IPSET_ATTR_TIMEOUT", timeout]]
        if bytes is not None:
            data_attrs += [["IPSET_ATTR_BYTES", bytes]]
        if packets is not None:
            data_attrs += [["IPSET_ATTR_PACKETS", packets]]
        if skbmark is not None:
            data_attrs += [["IPSET_ATTR_SKBMARK", skbmark]]
        if skbprio is not None:
            data_attrs += [["IPSET_ATTR_SKBPRIO", skbprio]]
        if skbqueue is not None:
            data_attrs += [["IPSET_ATTR_SKBQUEUE", skbqueue]]
        if adt_flags:
            data_attrs += [["IPSET_ATTR_CADT_FLAGS", adt_flags]]
        msg = ipset_msg()
        msg['attrs'] = [
            ['IPSET_ATTR_PROTOCOL', self._proto_version],
            ['IPSET_ATTR_SETNAME', name],
            ['IPSET_ATTR_DATA', {'attrs': data_attrs}],
        ]

        return self.request(
            msg,
            cmd,
            msg_flags=NLM_F_REQUEST | NLM_F_ACK | excl_flag,
            terminate=_nlmsg_error,
        )

    def add(
        self,
        name,
        entry,
        family=socket.AF_INET,
        exclusive=True,
        comment=None,
        timeout=None,
        etype="ip",
        skbmark=None,
        skbprio=None,
        skbqueue=None,
        wildcard=False,
        **kwargs
    ):
        '''
        Add a member to the ipset.

        etype is the entry type that you add to the ipset. It's related to
        the ipset type. For example, use "ip" for one hash:ip or bitmap:ip
        ipset.

        When your ipset store a tuple, like "hash:net,iface", you must use a
        comma a separator (etype="net,iface")

        entry is a string for "ip" and "net" objects. For ipset with several
        dimensions, you must use a tuple (or a list) of objects.

        "port" type is specific, since you can use integer of specialized
        containers like :class:`PortEntry` and :class:`PortRange`

        Examples::

            ipset = IPSet()
            ipset.create("foo", stype="hash:ip")
            ipset.add("foo", "198.51.100.1", etype="ip")

            ipset = IPSet()
            ipset.create("bar", stype="bitmap:port",
                         bitmap_ports_range=(1000, 2000))
            ipset.add("bar", 1001, etype="port")
            ipset.add("bar", PortRange(1500, 2000), etype="port")

            ipset = IPSet()
            import socket
            protocol = socket.getprotobyname("tcp")
            ipset.create("foobar", stype="hash:net,port")
            port_entry = PortEntry(80, protocol=protocol)
            ipset.add("foobar", ("198.51.100.0/24", port_entry),
                      etype="net,port")

        wildcard option enable kernel wildcard matching on interface
        name for net,iface entries.
        '''
        return self._add_delete_test(
            name,
            entry,
            family,
            IPSET_CMD_ADD,
            exclusive,
            comment=comment,
            timeout=timeout,
            etype=etype,
            skbmark=skbmark,
            skbprio=skbprio,
            skbqueue=skbqueue,
            wildcard=wildcard,
            **kwargs
        )

    def delete(
        self, name, entry, family=socket.AF_INET, exclusive=True, etype="ip"
    ):
        '''
        Delete a member from the ipset.

        See :func:`add` method for more information on etype.
        '''
        return self._add_delete_test(
            name, entry, family, IPSET_CMD_DEL, exclusive, etype=etype
        )

    def test(self, name, entry, family=socket.AF_INET, etype="ip"):
        '''
        Test if entry is part of an ipset

        See :func:`add` method for more information on etype.
        '''
        try:
            self._add_delete_test(
                name, entry, family, IPSET_CMD_TEST, False, etype=etype
            )
            return True
        except IPSetError as e:
            if e.code == IPSET_ERR_EXIST:
                return False
            raise e

    def swap(self, set_a, set_b):
        '''
        Swap two ipsets. They must have compatible content type.
        '''
        msg = ipset_msg()
        msg['attrs'] = [
            ['IPSET_ATTR_PROTOCOL', self._proto_version],
            ['IPSET_ATTR_SETNAME', set_a],
            ['IPSET_ATTR_TYPENAME', set_b],
        ]
        return self.request(
            msg,
            IPSET_CMD_SWAP,
            msg_flags=NLM_F_REQUEST | NLM_F_ACK,
            terminate=_nlmsg_error,
        )

    def flush(self, name=None):
        '''
        Flush all ipsets. When name is set, flush only this ipset.
        '''
        msg = ipset_msg()
        msg['attrs'] = [['IPSET_ATTR_PROTOCOL', self._proto_version]]
        if name is not None:
            msg['attrs'].append(['IPSET_ATTR_SETNAME', name])
        return self.request(
            msg,
            IPSET_CMD_FLUSH,
            msg_flags=NLM_F_REQUEST | NLM_F_ACK,
            terminate=_nlmsg_error,
        )

    def rename(self, name_src, name_dst):
        '''
        Rename the ipset.
        '''
        msg = ipset_msg()
        msg['attrs'] = [
            ['IPSET_ATTR_PROTOCOL', self._proto_version],
            ['IPSET_ATTR_SETNAME', name_src],
            ['IPSET_ATTR_TYPENAME', name_dst],
        ]
        return self.request(
            msg,
            IPSET_CMD_RENAME,
            msg_flags=NLM_F_REQUEST | NLM_F_ACK,
            terminate=_nlmsg_error,
        )

    def _get_set_by(self, cmd, value):
        # Check that IPSet version is supported
        if self._proto_version < 7:
            raise NotImplementedError()

        msg = ipset_msg()
        if cmd == IPSET_CMD_GET_BYNAME:
            msg['attrs'] = [
                ['IPSET_ATTR_PROTOCOL', self._proto_version],
                ['IPSET_ATTR_SETNAME', value],
            ]

        if cmd == IPSET_CMD_GET_BYINDEX:
            msg['attrs'] = [
                ['IPSET_ATTR_PROTOCOL', self._proto_version],
                ['IPSET_ATTR_INDEX', value],
            ]
        return self.request(msg, cmd)

    def get_set_byname(self, name):
        '''
        Get a set by its name
        '''

        return self._get_set_by(IPSET_CMD_GET_BYNAME, name)

    def get_set_byindex(self, index):
        '''
        Get a set by its index
        '''

        return self._get_set_by(IPSET_CMD_GET_BYINDEX, index)

    def get_supported_revisions(self, stype, family=socket.AF_INET):
        '''
        Return minimum and maximum of revisions supported by the kernel.

        Each ipset module (like hash:net, hash:ip, etc) has several
        revisions. Newer revisions often have more features or more
        performances. Thanks to this call, you can ask the kernel
        the list of supported revisions.

        You can manually set/force revisions used in IPSet constructor.

        Example::

            ipset = IPSet()
            ipset.get_supported_revisions("hash:net")

            ipset.get_supported_revisions("hash:net,port,net")
        '''
        msg = ipset_msg()
        msg['attrs'] = [
            ['IPSET_ATTR_PROTOCOL', self._proto_version],
            ['IPSET_ATTR_TYPENAME', stype],
            ['IPSET_ATTR_FAMILY', family],
        ]
        response = self.request(
            msg,
            IPSET_CMD_TYPE,
            msg_flags=NLM_F_REQUEST | NLM_F_ACK,
            terminate=_nlmsg_error,
        )

        min_revision = response[0].get_attr("IPSET_ATTR_PROTOCOL_MIN")
        max_revision = response[0].get_attr("IPSET_ATTR_REVISION")
        return min_revision, max_revision


class _IPSetError(IPSetError):
    '''
    Proxy class to not import all specifics ipset code in exceptions.py

    Out of the ipset module, a caller should use parent class instead
    '''

    def __init__(self, code, msg=None, cmd=None):
        if code in self.base_map:
            msg = self.base_map[code]
        elif cmd in self.cmd_map:
            error_map = self.cmd_map[cmd]
            if code in error_map:
                msg = error_map[code]
        super(_IPSetError, self).__init__(code, msg)

    base_map = {
        IPSET_ERR_PROTOCOL: "Kernel error received:" " ipset protocol error",
        IPSET_ERR_INVALID_CIDR: "The value of the CIDR parameter of"
        " the IP address is invalid",
        IPSET_ERR_TIMEOUT: "Timeout cannot be used: set was created"
        " without timeout support",
        IPSET_ERR_IPADDR_IPV4: "An IPv4 address is expected, but"
        " not received",
        IPSET_ERR_IPADDR_IPV6: "An IPv6 address is expected, but"
        " not received",
        IPSET_ERR_COUNTER: "Packet/byte counters cannot be used:"
        " set was created without counter support",
        IPSET_ERR_COMMENT: "Comment string is too long!",
        IPSET_ERR_SKBINFO: "Skbinfo mapping cannot be used: "
        " set was created without skbinfo support",
    }

    c_map = {
        errno.EEXIST: "Set cannot be created: set with the same"
        " name already exists",
        IPSET_ERR_FIND_TYPE: "Kernel error received: "
        "set type not supported",
        IPSET_ERR_MAX_SETS: "Kernel error received: maximal number of"
        " sets reached, cannot create more.",
        IPSET_ERR_INVALID_NETMASK: "The value of the netmask parameter"
        " is invalid",
        IPSET_ERR_INVALID_MARKMASK: "The value of the markmask parameter"
        " is invalid",
        IPSET_ERR_INVALID_FAMILY: "Protocol family not supported by the"
        " set type",
    }

    destroy_map = {
        IPSET_ERR_BUSY: "Set cannot be destroyed: it is in use"
        " by a kernel component"
    }

    r_map = {
        IPSET_ERR_EXIST_SETNAME2: "Set cannot be renamed: a set with the"
        " new name already exists",
        IPSET_ERR_REFERENCED: "Set cannot be renamed: it is in use by"
        " another system",
    }

    s_map = {
        IPSET_ERR_EXIST_SETNAME2: "Sets cannot be swapped: the second set"
        " does not exist",
        IPSET_ERR_TYPE_MISMATCH: "The sets cannot be swapped: their type"
        " does not match",
    }

    a_map = {
        IPSET_ERR_EXIST: "Element cannot be added to the set: it's"
        " already added"
    }

    del_map = {
        IPSET_ERR_EXIST: "Element cannot be deleted from the set:"
        " it's not added"
    }

    cmd_map = {
        IPSET_CMD_CREATE: c_map,
        IPSET_CMD_DESTROY: destroy_map,
        IPSET_CMD_RENAME: r_map,
        IPSET_CMD_SWAP: s_map,
        IPSET_CMD_ADD: a_map,
        IPSET_CMD_DEL: del_map,
    }



import os

import pyrad.packet
from pyrad.client import Client
from pyrad.dictionary import Dictionary


class RadiusAuthManager(object):
    def __init__(self, headers):
        user = headers['X-Auth-User']
        password = headers['X-Auth-Password']

        client = Client(
            server=os.environ.get('RADIUS_SERVER'),
            secret=os.environ.get('RADIUS_SECRET').encode('ascii'),
            dict=Dictionary('dictionary'),
        )

        req = client.CreateAuthPacket(
            code=pyrad.packet.AccessRequest, User_Name=user
        )

        req['User-Password'] = req.PwCrypt(password)
        reply = client.SendPacket(req)
        self.auth = reply.code

    def check(self, obj, tag):
        return self.auth == pyrad.packet.AccessAccept



from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg, protinfo_bridge

from .common import Index, IPRouteFilter, NLAKeyTransform


class BridgeFieldFilter(Index, NLAKeyTransform):
    _nla_prefix = ifinfmsg.prefix


class BridgeIPRouteFilter(IPRouteFilter):
    def build_vlan_info_spec(self, orig_spec):
        range_vids = [int(i) for i in str(orig_spec['vid']).split('-')]
        if len(range_vids) == 2:
            if 0 < int(range_vids[0]) < range_vids[1] < 4095:
                new_spec = []
                new_spec.append(
                    {
                        'vid': range_vids[0],
                        'flags': self.convert_flags('range_begin'),
                    }
                )
                new_spec.append(
                    {
                        'vid': range_vids[1],
                        'flags': self.convert_flags('range_end'),
                    }
                )
                return new_spec
        elif len(range_vids) == 1:
            if 0 < range_vids[0] < 4095:
                # PVID?
                if 'pvid' in orig_spec.keys():
                    if orig_spec['pvid']:
                        orig_spec['flags'] = self.convert_flags(
                            ['pvid', 'untagged']
                        )
                        del orig_spec['pvid']
                # Make sure the vid is an int.
                orig_spec['vid'] = range_vids[0]
                return [orig_spec]
        return []

    def build_vlan_tunnel_info_spec(self, orig_spec):
        # vlan_tunnel_info uses the same format as vlan_info,
        # just adds tunnel_id.
        vlan_info_spec = self.build_vlan_info_spec(orig_spec)
        range_ids = [int(i) for i in str(orig_spec['id']).split('-')]
        if len(range_ids) == 2 and len(vlan_info_spec) == 2:
            if 0 < range_ids[0] < range_ids[1] < 16777215:
                # vid to id mapping range must be the same length
                if (
                    vlan_info_spec[1]['vid'] - vlan_info_spec[0]['vid']
                    == range_ids[1] - range_ids[0]
                ):
                    vlan_info_spec[0]['id'] = range_ids[0]
                    vlan_info_spec[1]['id'] = range_ids[1]
                    return [
                        self.create_nla_spec(vlan_info_spec[0]),
                        self.create_nla_spec(vlan_info_spec[1]),
                    ]
        elif len(range_ids) == 1 and len(vlan_info_spec) == 1:
            if 0 < range_ids[0] < 4095:
                vlan_info_spec[0]['id'] = range_ids[0]
                # Delete flags because vlan_tunnel_info doesn't seem
                #  to use them, except for the RANGE.
                try:
                    del vlan_info_spec[0]['flags']
                except KeyError:
                    pass
                return [self.create_nla_spec(vlan_info_spec[0])]
        return []

    def create_nla_spec(self, spec):
        attrs = []
        for key in spec.keys():
            nla = ifinfmsg.af_spec_bridge.vlan_tunnel_info.name2nla(key)
            attrs.append([nla, spec[key]])
        return {'attrs': attrs}

    def convert_flags(self, flags):
        if isinstance(flags, int):
            return flags
        elif isinstance(flags, str):
            return ifinfmsg.af_spec_bridge.vlan_info.names2flags([flags])
        elif isinstance(flags, list):
            return ifinfmsg.af_spec_bridge.vlan_info.names2flags(flags)
        return 0

    def build_spec(self, orig_spec):
        if 'vid' in orig_spec.keys():
            if 'id' in orig_spec.keys():
                return self.build_vlan_tunnel_info_spec(orig_spec)
            else:
                return self.build_vlan_info_spec(orig_spec)
        return []

    def finalize(self, context):
        if self.command != 'dump':
            if 'IFLA_AF_SPEC' not in context:
                context['IFLA_AF_SPEC'] = {'attrs': []}
            for key in ('vlan_info', 'vlan_tunnel_info'):
                if key in context:
                    nla = ifinfmsg.af_spec_bridge.name2nla(key)
                    new_spec = self.build_spec(context[key])
                    for spec in new_spec:
                        context['IFLA_AF_SPEC']['attrs'].append([nla, spec])
                    try:
                        del context[key]
                    except KeyError:
                        pass
            for key in ('mode', 'vlan_flags'):
                if key in context:
                    nla = ifinfmsg.af_spec_bridge.name2nla(key)
                    context['IFLA_AF_SPEC']['attrs'].append(
                        [nla, context[key]]
                    )
                    try:
                        del context[key]
                    except KeyError:
                        pass


class BridgePortFieldFilter(IPRouteFilter):
    _nla_prefix = ifinfmsg.prefix
    _allowed = [x[0] for x in protinfo_bridge.nla_map]
    _allowed.append('attrs')

    def finalize(self, context):
        keys = tuple(context.keys())
        context['attrs'] = []
        for key in keys:
            context['attrs'].append(
                (protinfo_bridge.name2nla(key), context[key])
            )



def ipdb_interfaces_view(ndb):
    '''Provide read-only interfaces view with IPDB layout.

    In addition to standard NDB fields provides some IPDB
    specific fields.

    The method returns a simple dict structure, no background
    updates or system changes are supported.

    Please open a ticket on the project page if you are
    missing any attribute used in your project:

    https://github.com/svinota/pyroute2/issues
    '''
    ret = {}

    for record in ndb.interfaces.dump():
        interface = record._as_dict()
        interface['ipdb_scope'] = 'system'
        interface['ipdb_priority'] = 0
        interface['ipaddr'] = tuple(
            (
                (x.address, x.prefixlen)
                for x in (
                    ndb.addresses.dump().select_records(index=record.index)
                )
            )
        )
        interface['ports'] = tuple(
            (
                x.index
                for x in (
                    ndb.interfaces.dump().select_records(master=record.index)
                )
            )
        )
        interface['neighbours'] = tuple(
            (
                x.dst
                for x in (
                    ndb.neighbours.dump().select_records(ifindex=record.index)
                )
            )
        )
        ret[record.ifname] = interface

    return ret



from pyroute2.netlink import NETLINK_CONNECTOR, nlmsg
from pyroute2.netlink.nlsocket import NetlinkSocket


class cn_msg(nlmsg):
    fields = (
        ('idx', 'I'),
        ('val', 'I'),
        ('seq', 'I'),
        ('ack', 'I'),
        ('len', 'H'),
        ('flags', 'H'),
    )


class ConnectorSocket(NetlinkSocket):
    def __init__(self, fileno=None):
        super().__init__(NETLINK_CONNECTOR)



'''
htb
+++

TODO: list parameters

An example with htb qdisc, lets assume eth0 == 2::

    #          u32 -->    +--> htb 1:10 --> sfq 10:0
    #          |          |
    #          |          |
    # eth0 -- htb 1:0 -- htb 1:1
    #          |          |
    #          |          |
    #          u32 -->    +--> htb 1:20 --> sfq 20:0

    eth0 = 2
    # add root queue 1:0
    ip.tc("add", "htb", eth0, 0x10000, default=0x200000)

    # root class 1:1
    ip.tc("add-class", "htb", eth0, 0x10001,
          parent=0x10000,
          rate="256kbit",
          burst=1024 * 6)

    # two branches: 1:10 and 1:20
    ip.tc("add-class", "htb", eth0, 0x10010,
          parent=0x10001,
          rate="192kbit",
          burst=1024 * 6,
          prio=1)
    ip.tc("add-class", "htb", eht0, 0x10020,
          parent=0x10001,
          rate="128kbit",
          burst=1024 * 6,
          prio=2)

    # two leaves: 10:0 and 20:0
    ip.tc("add", "sfq", eth0, 0x100000,
          parent=0x10010,
          perturb=10)
    ip.tc("add", "sfq", eth0, 0x200000,
          parent=0x10020,
          perturb=10)

    # two filters: one to load packets into 1:10 and the
    # second to 1:20
    ip.tc("add-filter", "u32", eth0,
          parent=0x10000,
          prio=10,
          protocol=socket.AF_INET,
          target=0x10010,
          keys=["0x0006/0x00ff+8", "0x0000/0xffc0+2"])
    ip.tc("add-filter", "u32", eth0,
          parent=0x10000,
          prio=10,
          protocol=socket.AF_INET,
          target=0x10020,
          keys=["0x5/0xf+0", "0x10/0xff+33"])
'''
from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import RTM_DELQDISC, RTM_NEWQDISC, TC_H_ROOT
from pyroute2.netlink.rtnl.tcmsg.common import (
    calc_xmittime,
    get_hz,
    get_rate,
    nla_plus_rtab,
    stats2,
)

parent = TC_H_ROOT


def get_class_parameters(kwarg):
    prio = kwarg.get('prio', 0)
    mtu = kwarg.get('mtu', 1600)
    mpu = kwarg.get('mpu', 0)
    overhead = kwarg.get('overhead', 0)
    quantum = kwarg.get('quantum', 0)
    rate = get_rate(kwarg.get('rate', None))
    ceil = get_rate(kwarg.get('ceil', 0)) or rate

    burst = (
        kwarg.get('burst', None)
        or kwarg.get('maxburst', None)
        or kwarg.get('buffer', None)
    )

    if rate is not None:
        if burst is None:
            burst = rate / get_hz() + mtu
        burst = calc_xmittime(rate, burst)

    cburst = (
        kwarg.get('cburst', None)
        or kwarg.get('cmaxburst', None)
        or kwarg.get('cbuffer', None)
    )

    if ceil is not None:
        if cburst is None:
            cburst = ceil / get_hz() + mtu
        cburst = calc_xmittime(ceil, cburst)

    return {
        'attrs': [
            [
                'TCA_HTB_PARMS',
                {
                    'buffer': burst,
                    'cbuffer': cburst,
                    'quantum': quantum,
                    'prio': prio,
                    'rate': rate,
                    'ceil': ceil,
                    'ceil_overhead': overhead,
                    'rate_overhead': overhead,
                    'rate_mpu': mpu,
                    'ceil_mpu': mpu,
                },
            ],
            ['TCA_HTB_RTAB', True],
            ['TCA_HTB_CTAB', True],
        ]
    }


def get_parameters(kwarg):
    rate2quantum = kwarg.get('r2q', 0xA)
    version = kwarg.get('version', 3)
    defcls = kwarg.get('default', 0x10)

    return {
        'attrs': [
            [
                'TCA_HTB_INIT',
                {
                    'debug': 0,
                    'defcls': defcls,
                    'direct_pkts': 0,
                    'rate2quantum': rate2quantum,
                    'version': version,
                },
            ]
        ]
    }


def fix_msg(msg, kwarg):
    if not kwarg:
        opts = get_parameters({})
        msg['attrs'].append(['TCA_OPTIONS', opts])


# The tokens and ctokens are badly defined in the kernel structure
# as unsigned int instead of signed int. (cf net/sched/sch_htb.c
# in linux source)
class stats(nla):
    fields = (
        ('lends', 'I'),
        ('borrows', 'I'),
        ('giants', 'I'),
        ('tokens', 'i'),
        ('ctokens', 'i'),
    )


class qdisc_stats2(stats2):
    nla_map = (
        ('TCA_STATS_UNSPEC', 'none'),
        ('TCA_STATS_BASIC', 'basic'),
        ('TCA_STATS_RATE_EST', 'rate_est'),
        ('TCA_STATS_QUEUE', 'queue'),
    )


class class_stats2(stats2):
    class stats_app(stats):
        pass


def stats2(msg, *argv, **kwarg):
    if msg['header']['type'] in (RTM_NEWQDISC, RTM_DELQDISC):
        return qdisc_stats2
    else:
        return class_stats2


class options(nla_plus_rtab):
    nla_map = (
        ('TCA_HTB_UNSPEC', 'none'),
        ('TCA_HTB_PARMS', 'htb_parms'),
        ('TCA_HTB_INIT', 'htb_glob'),
        ('TCA_HTB_CTAB', 'ctab'),
        ('TCA_HTB_RTAB', 'rtab'),
    )

    class htb_glob(nla):
        fields = (
            ('version', 'I'),
            ('rate2quantum', 'I'),
            ('defcls', 'I'),
            ('debug', 'I'),
            ('direct_pkts', 'I'),
        )

    class htb_parms(nla_plus_rtab.parms):
        fields = (
            ('rate_cell_log', 'B'),
            ('rate___reserved', 'B'),
            ('rate_overhead', 'H'),
            ('rate_cell_align', 'h'),
            ('rate_mpu', 'H'),
            ('rate', 'I'),
            ('ceil_cell_log', 'B'),
            ('ceil___reserved', 'B'),
            ('ceil_overhead', 'H'),
            ('ceil_cell_align', 'h'),
            ('ceil_mpu', 'H'),
            ('ceil', 'I'),
            ('buffer', 'I'),
            ('cbuffer', 'I'),
            ('quantum', 'I'),
            ('level', 'I'),
            ('prio', 'I'),
        )



'''
.. testsetup::

    from pyroute2 import NDB
    ndb = NDB(sources=[{'target': 'localhost', 'kind': 'IPMock'}])

.. testsetup:: netns

    from types import MethodType

    from pyroute2 import NDB

    ndb = NDB(sources=[{'target': 'localhost', 'kind': 'IPMock'}])

    def add_mock_netns(self, netns):
        return self.add_orig(target=netns, kind='IPMock', preset='netns')

    ndb.sources.add_orig = ndb.sources.add
    ndb.sources.add = MethodType(add_mock_netns, ndb.sources)

.. testcleanup:: *

    for key, value in tuple(globals().items()):
        if key.startswith('ndb') and hasattr(value, 'close'):
            value.close()

NDB is a high level network management module. IT allows to manage interfaces,
routes, addresses etc. of connected systems, containers and network
namespaces.

In a nutshell, NDB collects and aggregates netlink events in an SQL database,
provides Python objects to reflect the system state, and applies changes back
to the system. The database expects updates only from the sources, no manual
SQL updates are expected normally.

.. aafig::
    :scale: 80
    :textual:

        +----------------------------------------------------------------+
      +----------------------------------------------------------------+ |
    +----------------------------------------------------------------+ | |
    |                                                                | | |
    |                              kernel                            | |-+
    |                                                                |-+
    +----------------------------------------------------------------+
            |                      | ^                     | ^
            | `netlink events`     | |                     | |
            | `inotify events`     | |                     | |
            | `...`                | |                     | |
            v                      v |                     v |
     +--------------+        +--------------+        +--------------+
     |     source   |        |     source   |        |     source   |<--\\
     +--------------+        +--------------+        +--------------+   |
            |                       |                       |           |
            |                       |                       |           |
            \\-----------------------+-----------------------/           |
                                    |                                   |
              parsed netlink events | `NDB._event_queue`                |
                                    |                                   |
                                    v                                   |
                        +------------------------+                      |
                        | `NDB.__dbm__()` thread |                      |
                        +------------------------+                      |
                                    |                                   |
                                    v                                   |
                     +-----------------------------+                    |
                     | `NDB.schema.load_netlink()` |                    |
                     | `NDB.objects.*.load*()`     |                    |
                     +-----------------------------+                    |
                                    |                                   |
                                    v                                   |
                         +----------------------+                       |
                         |  SQL database        |                       |
                         |     `SQLite`         |                       |
                         |     `PostgreSQL`     |                       |
                         +----------------------+                       |
                                    |                                   |
                                    |                                   |
                                    V                                   |
                              +---------------+                         |
                            +---------------+ |                         |
                          +---------------+ | |  `RTNL_Object.apply()`  |
                          | NDB object:   | | |-------------------------/
                          |  `interface`  | | |
                          |  `address`    | | |
                          |  `route`      | |-+
                          |  `...`        |-+
                          +---------------+

.. container:: aafig-caption

    object names on the diagram are clickable

The goal of NDB is to provide an easy access to RTNL info and entities via
Python objects, like `pyroute2.ndb.objects.interface` (see also:
:ref:`ndbinterfaces`), `pyroute2.ndb.objects.route` (see also:
:ref:`ndbroutes`) etc. These objects do not
only reflect the system state for the time of their instantiation, but
continuously monitor the system for relevant updates. The monitoring is
done via netlink notifications, thus no polling. Also the objects allow
to apply changes back to the system and rollback the changes.

On the other hand it's too expensive to create Python objects for all the
available RTNL entities, e.g. when there are hundreds of interfaces and
thousands of routes. Thus NDB creates objects only upon request, when
the user calls `.create()` to create new objects or runs
`ndb.<view>[selector]` (e.g. `ndb.interfaces['eth0']`) to access an
existing object.

To list existing RTNL entities NDB uses objects of the class `RecordSet`
that `yield` individual `Record` objects for every entity (see also:
:ref:`ndbreports`). An object of the `Record` class is immutable, doesn't
monitor any updates, doesn't contain any links to other objects and essentially
behaves like a simple named tuple.

.. aafig::
    :scale: 80
    :textual:


      +---------------------+
      |                     |
      |                     |
      | `NDB() instance`    |
      |                     |
      |                     |
      +---------------------+
                 |
                 |
        +-------------------+
      +-------------------+ |
    +-------------------+ | |-----------+--------------------------+
    |                   | | |           |                          |
    |                   | | |           |                          |
    | `View()`          | | |           |                          |
    |                   | |-+           |                          |
    |                   |-+             |                          |
    +-------------------+               |                          |
                               +------------------+       +------------------+
                               |                  |       |                  |
                               |                  |       |                  |
                               | `.dump()`        |       | `.create()`      |
                               | `.summary()`     |       | `.__getitem__()` |
                               |                  |       |                  |
                               |                  |       |                  |
                               +------------------+       +------------------+
                                        |                           |
                                        |                           |
                                        v                           v
                              +-------------------+        +------------------+
                              |                   |      +------------------+ |
                              |                   |    +------------------+ | |
                              | `RecordSet()`     |    | `Interface()`    | | |
                              |                   |    | `Address()`      | | |
                              |                   |    | `Route()`        | | |
                              +-------------------+    | `Neighbour()`    | | |
                                        |              | `Rule()`         | |-+
                                        |              |  ...             |-+
                                        v              +------------------+
                                +-------------------+
                              +-------------------+ |
                            +-------------------+ | |
                            | `filter()`        | | |
                            | `select()`        | | |
                            | `transform()`     | | |
                            | `join()`          | |-+
                            |  ...              |-+
                            +-------------------+
                                        |
                                        v
                                +-------------------+
                              +-------------------+ |
                            +-------------------+ | |
                            |                   | | |
                            |                   | | |
                            | `Record()`        | | |
                            |                   | |-+
                            |                   |-+
                            +-------------------+

.. container:: aafig-caption

    object names on the diagram are clickable

Here are some simple NDB usage examples. More info see in the reference
documentation below.

Print all the interface names on the system, assume we have an NDB
instance `ndb`:

.. testcode::

    for interface in ndb.interfaces.dump():
        print(interface.ifname)

.. testoutput::

    lo
    eth0

Print the routing information in the CSV format:

.. testcode::

    for record in ndb.routes.summary().format('csv'):
        print(record)

.. testoutput::

    'target','tflags','table','ifname','dst','dst_len','gateway'
    'localhost',0,254,'eth0','',0,'192.168.122.1'
    'localhost',0,254,'eth0','192.168.122.0',24,
    'localhost',0,255,'lo','127.0.0.0',8,
    'localhost',0,255,'lo','127.0.0.1',32,
    'localhost',0,255,'lo','127.255.255.255',32,
    'localhost',0,255,'eth0','192.168.122.28',32,
    'localhost',0,255,'eth0','192.168.122.255',32,

.. note:: More on report filtering and formatting: :ref:`ndbreports`

Print IP addresses of interfaces in several network namespaces as:

.. testcode:: netns

    nslist = ['netns01',
              'netns02',
              'netns03']

    for nsname in nslist:
        ndb.sources.add(netns=nsname)

    report = ndb.addresses.summary()
    report.select_records(target=lambda x: x.startswith('netns'))
    report.select_fields('address', 'ifname', 'target')
    for line in report.format('json'):
        print(line)

.. testoutput:: netns

    [
        {
            "address": "127.0.0.1",
            "ifname": "lo",
            "target": "netns01"
        },
        {
            "address": "127.0.0.1",
            "ifname": "lo",
            "target": "netns02"
        },
        {
            "address": "127.0.0.1",
            "ifname": "lo",
            "target": "netns03"
        }
    ]

Add an IP address on an interface:

.. testcode::

    with ndb.interfaces['eth0'] as eth0:
        eth0.add_ip('10.0.0.1/24')
    # ---> <---  NDB waits until the address setup

Change an interface property:

.. testcode::

    with ndb.interfaces['eth0'] as eth0:
        eth0.set(
            state='up',
            address='00:11:22:33:44:55',
        )
    # ---> <---  NDB waits here for the changes to be applied
    #            the commit() is called automatically by the
    #            context manager's __exit__()

'''
import atexit
import ctypes
import ctypes.util
import logging
import logging.handlers
import sys
import threading

from pyroute2 import config
from pyroute2.common import basestring

##
# NDB stuff
from .auth_manager import AuthManager
from .events import ShutdownException
from .messages import cmsg
from .schema import DBProvider
from .task_manager import TaskManager
from .transaction import Transaction
from .view import SourcesView, View

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

try:
    import queue
except ImportError:
    import Queue as queue

log = logging.getLogger(__name__)


NDB_VIEWS_SPECS = (
    ('interfaces', 'interfaces'),
    ('addresses', 'addresses'),
    ('routes', 'routes'),
    ('neighbours', 'neighbours'),
    ('af_bridge_fdb', 'fdb'),
    ('rules', 'rules'),
    ('netns', 'netns'),
    ('af_bridge_vlans', 'vlans'),
)


class Log:
    def __init__(self, log_id=None):
        self.logger = None
        self.state = False
        self.log_id = log_id or id(self)
        self.logger = logging.getLogger('pyroute2.ndb.%s' % self.log_id)
        self.main = self.channel('main')

    def __call__(self, target=None, level=logging.INFO):
        if target is None:
            return self.logger is not None

        if self.logger is not None:
            for handler in tuple(self.logger.handlers):
                self.logger.removeHandler(handler)

        if target in ('off', False):
            if self.state:
                self.logger.setLevel(0)
                self.logger.addHandler(logging.NullHandler())
            return

        if target in ('on', 'stderr'):
            handler = logging.StreamHandler()
        elif target == 'debug':
            handler = logging.StreamHandler()
            level = logging.DEBUG
        elif isinstance(target, basestring):
            url = urlparse(target)
            if not url.scheme and url.path:
                handler = logging.FileHandler(url.path)
            elif url.scheme == 'syslog':
                handler = logging.handlers.SysLogHandler(
                    address=url.netloc.split(':')
                )
            else:
                raise ValueError('logging scheme not supported')
        else:
            handler = target

        # set formatting only for new created logging handlers
        if handler is not target:
            fmt = '%(asctime)s %(levelname)8s %(name)s: %(message)s'
            formatter = logging.Formatter(fmt)
            handler.setFormatter(formatter)

        self.logger.addHandler(handler)
        self.logger.setLevel(level)

    @property
    def on(self):
        self.__call__(target='on')

    @property
    def off(self):
        self.__call__(target='off')

    def close(self):
        manager = self.logger.manager
        name = self.logger.name
        # the loggerDict can be huge, so don't
        # cache all the keys -- cache only the
        # needed ones
        purge_list = []
        for logger in manager.loggerDict.keys():
            if logger.startswith(name):
                purge_list.append(logger)
        # now shoot them one by one
        for logger in purge_list:
            del manager.loggerDict[logger]
        # don't force GC, leave it to the user
        del manager
        del name
        del purge_list

    def channel(self, name):
        return logging.getLogger('pyroute2.ndb.%s.%s' % (self.log_id, name))

    def debug(self, *argv, **kwarg):
        return self.main.debug(*argv, **kwarg)

    def info(self, *argv, **kwarg):
        return self.main.info(*argv, **kwarg)

    def warning(self, *argv, **kwarg):
        return self.main.warning(*argv, **kwarg)

    def error(self, *argv, **kwarg):
        return self.main.error(*argv, **kwarg)

    def critical(self, *argv, **kwarg):
        return self.main.critical(*argv, **kwarg)


class DeadEnd:
    def put(self, *argv, **kwarg):
        raise ShutdownException('shutdown in progress')


class EventQueue:
    def __init__(self, *argv, **kwarg):
        self._bypass = self._queue = queue.Queue(*argv, **kwarg)

    def put(self, msg, source=None):
        return self._queue.put((source, msg))

    def shutdown(self):
        self._queue = DeadEnd()

    def bypass(self, msg, source=None):
        return self._bypass.put((source, msg))

    def get(self, *argv, **kwarg):
        return self._bypass.get(*argv, **kwarg)

    def qsize(self):
        return self._bypass.qsize()


class AuthProxy:
    def __init__(self, ndb, auth_managers):
        self._ndb = ndb
        self._auth_managers = auth_managers

        for vtable, vname in NDB_VIEWS_SPECS:
            view = View(self._ndb, vtable, auth_managers=self._auth_managers)
            setattr(self, vname, view)


class NDB:
    @property
    def nsmanager(self):
        return '%s/nsmanager' % self.localhost

    def __init__(
        self,
        sources=None,
        localhost='localhost',
        db_provider='sqlite3',
        db_spec=':memory:',
        db_cleanup=True,
        rtnl_debug=False,
        log=False,
        auto_netns=False,
        libc=None,
    ):
        if db_provider == 'postgres':
            db_provider = 'psycopg2'

        self.localhost = localhost
        self.schema = None
        self.libc = libc or ctypes.CDLL(
            ctypes.util.find_library('c'), use_errno=True
        )
        self.log = Log(log_id=id(self))
        self._db = None
        self._dbm_thread = None
        self._dbm_ready = threading.Event()
        self._dbm_shutdown = threading.Event()
        self._global_lock = threading.Lock()
        self._event_queue = EventQueue(maxsize=100)
        self.messenger = None
        #
        if log:
            if isinstance(log, basestring):
                self.log(log)
            elif isinstance(log, (tuple, list)):
                self.log(*log)
            elif isinstance(log, dict):
                self.log(**log)
            else:
                raise TypeError('wrong log spec format')
        #
        # fix sources prime
        if sources is None:
            if config.mock_iproute:
                sources = [{'target': 'localhost', 'kind': 'IPMock'}]
            else:
                sources = [
                    {
                        'target': self.localhost,
                        'kind': 'local',
                        'nlm_generator': 1,
                    }
                ]
                if sys.platform.startswith('linux'):
                    sources.append(
                        {'target': self.nsmanager, 'kind': 'nsmanager'}
                    )
        elif not isinstance(sources, (list, tuple)):
            raise ValueError('sources format not supported')

        for spec in sources:
            if 'target' not in spec:
                spec['target'] = self.localhost
                break

        am = AuthManager(
            {'obj:list': True, 'obj:read': True, 'obj:modify': True},
            self.log.channel('auth'),
        )
        self.sources = SourcesView(self, auth_managers=[am])
        self._call_registry = {}
        self._nl = sources
        atexit.register(self.close)
        self._dbm_ready.clear()
        self._dbm_error = None
        self.config = {
            'provider': str(DBProvider(db_provider)),
            'spec': db_spec,
            'rtnl_debug': rtnl_debug,
            'db_cleanup': db_cleanup,
            'auto_netns': auto_netns,
        }
        self.task_manager = TaskManager(self)
        self._dbm_thread = threading.Thread(
            target=self.task_manager.run, name='NDB main loop'
        )
        self._dbm_thread.daemon = True
        self._dbm_thread.start()
        self._dbm_ready.wait()
        if self._dbm_error is not None:
            raise self._dbm_error
        for vtable, vname in NDB_VIEWS_SPECS:
            view = View(self, vtable, auth_managers=[am])
            setattr(self, vname, view)
        # self.query = Query(self.schema)

    def _get_view(self, table, chain=None, auth_managers=None):
        return View(self, table, chain, auth_managers)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def begin(self):
        return Transaction(self.log.channel('transaction'))

    def readonly(self):
        return self.auth_proxy(
            AuthManager(
                {'obj:list': True, 'obj:read': True, 'obj:modify': False},
                self.log.channel('auth'),
            )
        )

    def auth_proxy(self, auth_manager):
        return AuthProxy(self, [auth_manager])

    def close(self):
        with self._global_lock:
            if self._dbm_shutdown.is_set():
                return
            else:
                self._dbm_shutdown.set()
            if hasattr(atexit, 'unregister'):
                atexit.unregister(self.close)
            else:
                try:
                    atexit._exithandlers.remove((self.close, (), {}))
                except ValueError:
                    pass
            # shutdown the _dbm_thread
            self._event_queue.shutdown()
            self._event_queue.bypass((cmsg(None, ShutdownException()),))
            self._dbm_thread.join()
            # shutdown the logger -- free the resources
            self.log.close()

    def backup(self, spec):
        self.task_manager.db_backup(spec)

    def reload(self, kinds=None):
        for source in self.sources.values():
            if kinds is not None and source.kind in kinds:
                source.restart()



'''

.. testsetup:: *

    from socket import AF_INET
    from pyroute2 import NDB
    from pyroute2 import config
    config.mock_iproute = True
    ndb = NDB()

.. testcleanup:: *

    ndb.close()

Using the global view
=====================

The `addresses` view provides access to all the addresses registered in the DB,
as well as methods to create and remove them:

.. testcode::

    eth0 = ndb.interfaces['eth0']

    # create an address
    ndb.addresses.create(
        address='10.0.0.1',
        prefixlen=24,
        index=eth0['index'],
    ).commit()

    # remove it
    with ndb.addresses['10.0.0.1/24'] as addr:
        addr.remove()

    # list addresses
    for record in ndb.addresses.summary():
        print(record)

.. testoutput::
    :hide:

    ('localhost', 0, 'lo', '127.0.0.1', 8)
    ('localhost', 0, 'eth0', '192.168.122.28', 24)


Using ipaddr views
==================

Interfaces also provide address views as subsets of the global
address view:

.. testcode::

    with ndb.interfaces['eth0'] as eth0:
        for record in eth0.ipaddr.summary():
            print(record)

.. testoutput::

    ('localhost', 0, 'eth0', '192.168.122.28', 24)

It is possible use the same API as with the global address view:

.. testcode::

    with ndb.interfaces['eth0'] as eth0:
        eth0.ipaddr.create(
            address='10.0.0.1', prefixlen=24  # index is implied
        ).commit()
    for record in ndb.addresses.summary():
        print(record)

.. testoutput::

    ('localhost', 0, 'lo', '127.0.0.1', 8)
    ('localhost', 0, 'eth0', '10.0.0.1', 24)
    ('localhost', 0, 'eth0', '192.168.122.28', 24)

Using interface methods
=======================

Interfaces provide also simple methods to manage addresses:

.. testcode::

    with ndb.interfaces['eth0'] as eth0:
        eth0.del_ip('192.168.122.28/24')  # remove an existing address
        eth0.del_ip(family=AF_INET)  # ... or remove all IPv4 addresses
        eth0.add_ip('10.0.0.1/24')  # add a new IP address
        eth0.add_ip(address='10.0.0.2', prefixlen=24)  # ... or using keywords
        eth0.set('state', 'up')
    with ndb.addresses.summary() as report:
        report.select_records(ifname='eth0')
        for address in report:
            print(address)

.. testoutput::

    ('localhost', 0, 'eth0', '10.0.0.1', 24)
    ('localhost', 0, 'eth0', '10.0.0.2', 24)


Functions `add_ip()` and `del_ip()` return the interface object, so they
can be chained as in the example above, and the final `commit()` will
commit all the changes in the chain.

The keywords to `del_ip()` are the same object field names that may be used
in the selectors or report filters:

.. testcode::

    with ndb.interfaces['eth0'] as eth0:
        eth0.del_ip(prefixlen=24)  # remove all addresses with mask /24

A match function that may be passed to the `del_ip()` is the same as for
`addresses.dump().select_records()`, and it gets a named tuple as the argument.
The fields are named in the same way as address objects fields. So if you
want to filter addresses by a pattern or the `prefixlen` field with a
match function, you may use:

.. testcode:: x1

    with ndb.interfaces['eth0'] as eth0:
        eth0.add_ip('10.0.0.1/25')

    with ndb.interfaces['eth0'] as eth0:
        eth0.del_ip(lambda x: x.address.startswith('192.168'))
        eth0.del_ip(lambda x: x.prefixlen == 25)

An empty `del_ip()` removes all the IP addresses on the interface:

.. testcode:: x2

    with ndb.interfaces['eth0'] as eth0:
        eth0.del_ip()  # flush all the IP:s

Accessing one address details
=============================

Access an address as a separate RTNL object:

.. testcode:: x3

    print(ndb.addresses['192.168.122.28/24'])

.. testoutput:: x3
    :hide:

    {'target': 'localhost', 'address': '192.168.122.28', 'prefixlen': 24, \
'tflags': 0, 'family': 2, 'index': 2, 'local': '192.168.122.28', \
'flags': 512, 'scope': 0, 'label': 'eth0', 'broadcast': '192.168.122.255', \
'anycast': None, 'multicast': None}

Please notice that address objects are read-only, you may not change them,
only remove old ones, and create new.
'''
from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg
from pyroute2.requests.address import AddressFieldFilter

from ..objects import RTNL_Object


def load_ifaddrmsg(schema, target, event):
    #
    # bypass
    #
    schema.load_netlink('addresses', target, event)
    #
    # last address removal should trigger routes flush
    # Bug-Url: https://github.com/svinota/pyroute2/issues/849
    #
    if event['header']['type'] % 2 and event.get('index'):
        #
        # check IPv4 addresses on the interface
        #
        addresses = schema.execute(
            '''
                              SELECT * FROM addresses WHERE
                              f_target = %s AND
                              f_index = %s AND
                              f_family = 2
                              '''
            % (schema.plch, schema.plch),
            (target, event['index']),
        ).fetchmany()
        if not len(addresses):
            schema.execute(
                '''
                           DELETE FROM routes WHERE
                           f_target = %s AND
                           f_RTA_OIF = %s OR
                           f_RTA_IIF = %s
                           '''
                % (schema.plch, schema.plch, schema.plch),
                (target, event['index'], event['index']),
            )


ifaddr_spec = (
    ifaddrmsg.sql_schema()
    .unique_index('family', 'prefixlen', 'index', 'IFA_ADDRESS', 'IFA_LOCAL')
    .foreign_key(
        'interfaces',
        ('f_target', 'f_tflags', 'f_index'),
        ('f_target', 'f_tflags', 'f_index'),
    )
)

init = {
    'specs': [['addresses', ifaddr_spec]],
    'classes': [['addresses', ifaddrmsg]],
    'event_map': {ifaddrmsg: [load_ifaddrmsg]},
}


class Address(RTNL_Object):
    table = 'addresses'
    msg_class = ifaddrmsg
    field_filter = AddressFieldFilter
    api = 'addr'

    @classmethod
    def _count(cls, view):
        if view.chain:
            return view.ndb.task_manager.db_fetchone(
                'SELECT count(*) FROM %s WHERE f_index = %s'
                % (view.table, view.ndb.schema.plch),
                [view.chain['index']],
            )
        else:
            return view.ndb.task_manager.db_fetchone(
                'SELECT count(*) FROM %s' % view.table
            )

    @classmethod
    def _dump_where(cls, view):
        if view.chain:
            plch = view.ndb.schema.plch
            where = '''
                    WHERE
                        main.f_target = %s AND
                        main.f_index = %s
                    ''' % (
                plch,
                plch,
            )
            values = [view.chain['target'], view.chain['index']]
        else:
            where = ''
            values = []
        return (where, values)

    @classmethod
    def summary(cls, view):
        req = '''
              SELECT
                  main.f_target, main.f_tflags,
                  intf.f_IFLA_IFNAME, main.f_IFA_ADDRESS, main.f_prefixlen
              FROM
                  addresses AS main
              INNER JOIN
                  interfaces AS intf
              ON
                  main.f_index = intf.f_index
                  AND main.f_target = intf.f_target
              '''
        yield ('target', 'tflags', 'ifname', 'address', 'prefixlen')
        where, values = cls._dump_where(view)
        for record in view.ndb.task_manager.db_fetch(req + where, values):
            yield record

    def mark_tflags(self, mark):
        plch = (self.schema.plch,) * 3
        self.schema.execute(
            '''
                            UPDATE interfaces SET
                                f_tflags = %s
                            WHERE f_index = %s AND f_target = %s
                            '''
            % plch,
            (mark, self['index'], self['target']),
        )

    def __init__(self, *argv, **kwarg):
        kwarg['iclass'] = ifaddrmsg
        self.event_map = {ifaddrmsg: "load_rtnlmsg"}
        super(Address, self).__init__(*argv, **kwarg)

    @staticmethod
    def compare_record(left, right):
        if isinstance(right, str):
            return right == left['address'] or right == '%s/%i' % (
                left['address'],
                left['prefixlen'],
            )

    @classmethod
    def spec_normalize(cls, processed, spec):
        '''
        Address key normalization::

            { ... }        ->  { ... }
            "10.0.0.1/24"  ->  {"address": "10.0.0.1",
                                "prefixlen": 24}
        '''
        if isinstance(spec, str):
            processed['address'] = spec
        return processed

    def key_repr(self):
        return '%s/%s %s/%s' % (
            self.get('target', ''),
            self.get('label', self.get('index', '')),
            self.get('local', self.get('address', '')),
            self.get('prefixlen', ''),
        )
